#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import pytest
from fixtures import *
from helpers import *
from pywr._core import Timestep, ScenarioIndex
import pandas
from pywr.core import *
from pywr.domains.river import *
from pywr.parameters import Parameter, ConstantParameter, DataFrameParameter, AggregatedParameter
from pywr.recorders import assert_rec, AssertionRecorder

TEST_FOLDER = os.path.dirname(__file__)

def test_names():
    '''Test node names'''
    model = Model()

    node1 = Input(model, name='A')
    node2 = Output(model, name='B')
    assert(model.nodes['A'] is node1)
    assert(model.nodes['B'] is node2)

    nodes = sorted(model.nodes, key=lambda node: node.name)
    assert(nodes == [node1, node2])

    # rename node
    node1.name = 'C'
    assert(model.nodes['C'] is node1)
    assert('A' not in model.nodes)

    # attempt name collision (via rename)
    with pytest.raises(ValueError):
        node2.name = 'C'

    # attempt name collision (via new)
    with pytest.raises(ValueError):
        node3 = Input(model, name='C')
    assert(len(model.nodes) == 2)  # node3 not added to graph

    # attempt to create a node without a name
    with pytest.raises(TypeError):
        node4 = Input(model)


def test_model_nodes(model):
    """Test Model.nodes API"""
    node = Input(model, 'test')

    # test node by index
    assert(model.nodes['test'] is node)

    with pytest.raises(KeyError):
        model.nodes['invalid']

    # test node iterator
    all_nodes = [node for node in model.nodes]
    assert(all_nodes == [node])

    # support for item deletion
    del(model.nodes['test'])
    all_nodes = [node for node in model.nodes]
    assert(all_nodes == [])


def test_unexpected_kwarg_node():
    model = Model()

    with pytest.raises(TypeError):
        node = Node(model, 'test_node', invalid=True)
    with pytest.raises(TypeError):
        inpt = Input(model, 'test_input', invalid=True)
    with pytest.raises(TypeError):
        storage = Storage(model, 'test_storage', invalid=True)
    # none of the nodes should have been added to the model as they all
    # raised exceptions during __init__
    assert(not model.nodes)


def test_unexpected_kwarg_model():
    with pytest.raises(TypeError):
        model = Model(thisisgoingtofail=True)
    model = Model()

def test_slots_connect_disconnect():
    """Test connection and disconnection to storage node slots
    """
    model = Model()

    supply1 = Input(model, name='supply1')
    supply2 = Input(model, name='supply2')
    storage = Storage(model, name='storage', inputs=2, outputs=2)

    storage_inputs = [[x for x in storage.iter_slots(slot_name=n, is_connector=True)][0] for n in (0, 1)]
    storage_outputs = [[x for x in storage.iter_slots(slot_name=n, is_connector=False)][0] for n in (0, 1)]

    # attempt to connect to invalid slot
    # an error is raised, and no connection is made
    with pytest.raises(IndexError):
        supply1.connect(storage, to_slot=3)
    assert((supply1, storage_inputs[0]) not in model.edges())
    assert((supply1, storage_inputs[1]) not in model.edges())

    # connect node to storage slot 0
    supply1.connect(storage, to_slot=0)
    assert((supply1, storage_outputs[0]) in model.edges())
    assert((supply1, storage_inputs[1]) not in model.edges())

    # the same node can be connected to multiple slots
    # connect node to storage slot 1
    supply1.connect(storage, to_slot=1)
    assert((supply1, storage_outputs[0]) in model.edges())
    assert((supply1, storage_outputs[1]) in model.edges())

    # disconnect the node from a particular slot (recommended method)
    supply1.disconnect(storage, slot_name=0)
    assert((supply1, storage_outputs[0]) not in model.edges())
    assert((supply1, storage_outputs[1]) in model.edges())

    # disconnect the node from a particular slot (direct, not recommended)
    supply1.connect(storage, to_slot=0)
    supply1.disconnect(storage_outputs[0])
    assert((supply1, storage_outputs[0]) not in model.edges())
    assert((supply1, storage_outputs[1]) in model.edges())

    # specifying the storage in general removes the connection from all slots
    supply1.connect(storage, to_slot=0)
    supply1.disconnect(storage)
    assert((supply1, storage_outputs[0]) not in model.edges())
    assert((supply1, storage_outputs[1]) not in model.edges())

    # it's an error to attempt to disconnect if nodes aren't connected
    with pytest.raises(Exception):
        supply1.disconnect(storage)


def test_node_position():
    model = Model()

    # node position, from kwargs

    node1 = Input(model, "input", position={"schematic": (10, 20), "geographic": (-1, 52)})

    # node position, from JSON

    data = {
        "name": "output",
        "position": {
            "schematic": (30, 40),
            "geographic": (-1.5, 52.2),
        }
    }
    node2 = Output.pre_load(model, data)
    node2.finalise_load()

    assert(node1.position["schematic"] == (10, 20))
    assert(node1.position["geographic"] == (-1, 52))
    assert(node2.position["schematic"] == (30, 40))
    assert(node2.position["geographic"] == (-1.5, 52.2))

    node1.position["schematic"] = (50, 60)
    assert(node1.position["schematic"] == (50, 60))

    # node without position

    node3 = Node(model, "node3")
    assert(node3.position == {})

    # reservoir position, from JSON

    data = {
        "name": "reservoir",
        "position": {
            "schematic": (99, 70),
            "geographic": (-2.5, 55.6),
        },
        "max_volume": 1000,
        "initial_volume": 500
    }

    storage = Storage.pre_load(model, data)
    storage.finalise_load()

    assert(storage.position["schematic"] == (99, 70))
    assert(storage.position["geographic"] == (-2.5, 55.6))


test_data = ["timeseries1.xlsx", os.path.join("models", "timeseries1.csv")]
@pytest.mark.parametrize("filename", test_data)
def test_timeseries_excel(simple_linear_model, filename):
    """Test creation of a DataFrameParameter from external data (e.g. CSV)"""
    model = simple_linear_model

    # create DataFrameParameter from external data
    filename = os.path.join(TEST_FOLDER, filename)
    data = {"url": filename, "column": "Data", "index_col": "Timestamp"}
    if filename.endswith(".csv"):
        data.update({"parse_dates": True, "dayfirst":True})
    ts = DataFrameParameter.load(model, data)

    # model (intentionally not aligned)
    index = ts.dataframe.index
    model.timestepper.start = index[0] + (5 * index.freq)
    model.timestepper.end = index[-1] - (12 * index.freq)

    # need to assign parameter for it's setup method to be called
    model.nodes["Input"].max_flow = ts

    @assert_rec(model, ts)
    def expected(timestep, scenario_index):
        return ts.dataframe.loc[timestep.datetime]

    model.run()

def test_dirty_model():
    """Test that the LP is updated when the model structure is redefined"""
    # start dirty
    model = Model()
    assert(model.dirty)

    # add some nodes, still dirty
    supply1 = Input(model, 'supply1')
    demand1 = Output(model, 'demand1')
    supply1.connect(demand1)
    assert(model.dirty)

    # run the model, clean
    result = model.step()
    assert(not model.dirty)

    # add a new node, dirty
    supply2 = Input(model, 'supply2')

    # run the model, clean
    result = model.step()
    assert(not model.dirty)

    # add a new connection, dirty
    supply2.connect(demand1)
    assert(model.dirty)

    # run the model, clean
    result = model.step()
    assert(not model.dirty)

    # remove a connection, dirty
    supply2.disconnect()
    assert(model.dirty)

def test_reset_initial_volume():
    """
    If the model doesn't reset correctly changing the initial volume and
    re-running will cause an exception.
    """
    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01')
    )

    storage = Storage(model, 'storage', inputs=1, outputs=0)
    otpt = Output(model, 'output', max_flow=99999, cost=-99999)
    storage.connect(otpt)

    model.check()

    for initial_volume in (50, 100):
        storage.max_volume = initial_volume
        storage.initial_volume = initial_volume
        model.run()
        assert(otpt.flow == initial_volume)

def test_shorthand_property():
    # test shorthand assignment of constant properties
    model = Model()
    node = Node(model, 'node')
    for attr in ('min_flow', 'max_flow', 'cost', 'conversion_factor'):
        # should except int, float or Paramter
        setattr(node, attr, 123)
        if attr == 'conversion_factor':
            with pytest.raises(ValueError):
                setattr(node, attr, Parameter(model))
        else:
            setattr(node, attr, Parameter(model))

        with pytest.raises(TypeError):
            setattr(node, attr, '123')
            setattr(node, attr, None)


def test_shorthand_property_storage():
    # test shorthand assignment of constant properties
    model = Model()
    node = Storage(model, 'node')
    for attr in ('min_volume', 'max_volume', 'cost', 'level'):
        # should except int, float or Paramter
        setattr(node, attr, 123)
        if attr == 'conversion_factor':
            with pytest.raises(ValueError):
                setattr(node, attr, Parameter(model))
        else:
            setattr(node, attr, Parameter(model))

        with pytest.raises(TypeError):
            setattr(node, attr, '123')
            setattr(node, attr, None)


def test_reset_before_run():
    # See issue #82. Previously this would raise:
    #    AttributeError: Memoryview is not initialized
    model = Model()
    node = Node(model, 'node')
    model.reset()


def test_check_isolated_nodes(simple_linear_model):
    """Test model storage checker"""
    # the simple model shouldn't have any isolated nodes
    model = simple_linear_model
    model.check()

    # add a node, but don't connect it to the network
    isolated_node = Input(model, 'isolated')
    with pytest.raises(ModelStructureError):
        model.check()

def test_check_isolated_nodes_storage():
    """Test model structure checker with Storage

    The Storage node itself doesn't have any connections, but it's child
    nodes do need to be connected.
    """
    model = Model()

    # add a storage, but don't connect it's outflow to anything
    storage = Storage(model, 'storage', inputs=1, outputs=0, initial_volume=0.0)
    with pytest.raises(ModelStructureError):
        model.check()

    # add a demand node and connect it to the storage outflow
    demand = Output(model, 'demand')
    storage.connect(demand, from_slot=0)
    model.check()

def test_storage_max_volume_zero():
    """Test a that an max_volume of zero results in a NaN for current_pc and no exception

    """

    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01')
    )

    storage = Storage(model, 'storage', inputs=1, outputs=0, initial_volume=0.0)
    otpt = Output(model, 'output', max_flow=99999, cost=-99999)
    storage.connect(otpt)

    storage.max_volume = 0

    model.run()
    assert np.isnan(storage.current_pc)


def test_storage_max_volume_param():
    """Test a that an max_volume with a Parameter results in the correct current_pc

    """

    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01')
    )

    storage = Storage(model, 'storage', inputs=1, outputs=0)
    otpt = Output(model, 'output', max_flow=99999, cost=-99999)
    storage.connect(otpt)

    p = ConstantParameter(model, 20.0)
    storage.max_volume = p
    storage.initial_volume = 10.0
    storage.initial_volume_pc = 0.5

    model.setup()
    np.testing.assert_allclose(storage.current_pc, 0.5)

    model.run()

    p.set_double_variables(np.asarray([40.0, ]))
    model.reset()

    # This is a demonstration of the issue describe in #470
    #   https://github.com/pywr/pywr/issues/470
    # The initial storage is defined in both absolute and relative terms
    # but these are now not consistent with one another and the updated max_volume

    np.testing.assert_allclose(storage.volume, 10.0)
    np.testing.assert_allclose(storage.current_pc, 0.5)


def test_storage_initial_volume_pc():
    """Test that setting initial volume as a percentage works as expected.
    """
    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01')
    )

    storage = Storage(model, 'storage', inputs=1, outputs=0, initial_volume_pc=0.5, max_volume=20.0)
    otpt = Output(model, 'output', max_flow=99999, cost=-99999)
    storage.connect(otpt)

    model.setup()
    np.testing.assert_allclose(storage.current_pc, 0.5)
    np.testing.assert_allclose(storage.volume, 10.0)

    model.run()

    storage.max_volume = 40.0
    model.reset()
    np.testing.assert_allclose(storage.current_pc, 0.5)
    np.testing.assert_allclose(storage.volume, 20.0)


def test_storage_initial_missing_raises():
    """Test that a RuntimeError is raised if no initial volume is specified.
    """
    model = Model()

    storage = Storage(model, 'storage', inputs=1, outputs=0, max_volume=20.0)
    otpt = Output(model, 'output', max_flow=99999, cost=-99999)
    storage.connect(otpt)

    with pytest.raises(RuntimeError):
        model.run()


def test_storage_initial_volume_table():
    """ Test loading initial storage volume from a table """
    filename = os.path.join(TEST_FOLDER, "models", "reservoir_initial_vol_from_table.json")
    model = Model.load(filename)

    np.testing.assert_allclose(model.nodes["supply1"].initial_volume, 0.0)
    np.testing.assert_allclose(model.nodes["supply2"].initial_volume, 35.0)


def test_recursive_delete():
    """Test recursive deletion of child nodes for compound nodes"""
    model = Model()
    n1 = Input(model, "n1")
    n2 = Output(model, "n2")
    s = Storage(model, "s", outputs=2)
    assert len(model.nodes) == 3
    assert len(model.graph.nodes()) == 6
    del(model.nodes["n1"])
    assert len(model.nodes) == 2
    assert len(model.graph.nodes()) == 5
    del(model.nodes["s"])
    assert len(model.nodes) == 1
    assert len(model.graph.nodes()) == 1


def test_json_include():
    """Test include in JSON document"""
    filename = os.path.join(TEST_FOLDER, "models", "extra1.json")
    model = Model.load(filename)

    supply1 = model.nodes["supply1"]
    supply2 = model.nodes["supply2"]
    assert(isinstance(supply2.max_flow, ConstantParameter))


def test_py_include():
    """Test include in Python document"""
    filename = os.path.join(TEST_FOLDER, "models", "python_include.json")
    model = Model.load(filename)

    model.run()


def test_json_min_version():
    """Test warning is raised if document minimum version is more than we have"""
    filename = os.path.join(TEST_FOLDER, "models", "version1.json")
    with pytest.warns(RuntimeWarning):
        model = Model.load(filename)

def test_initial_timestep():
    """Current timestep before model has started is undefined"""
    filename = os.path.join(TEST_FOLDER, "models", "extra1.json")
    model = Model.load(filename)
    assert(model.timestepper.current is None)
    model.run()
    assert(isinstance(model.timestepper.current, Timestep))

def test_timestepper_repr(model):
    timestepper = model.timestepper
    print(timestepper)

def test_timestep_repr():
    filename = os.path.join(TEST_FOLDER, "models", "simple1.json")
    model = Model.load(filename)
    model.timestepper.end = "2015-01-05"
    res = model.run()
    assert(isinstance(res.timestep, Timestep))
    assert("2015-01-05" in str(res.timestep))

def test_virtual_storage_cost():
    """VirtualStorage doesn't (currently) implement its cost attribute"""
    model = Model()
    A = Input(model, "A")
    B = Output(model, "B")
    A.connect(B)
    node = VirtualStorage(model, "storage", [A, B])
    node.check()
    node.cost = 5.0
    with pytest.raises(NotImplementedError):
        model.check()

def test_json_invalid():
    """JSON exceptions should report file name"""
    filename = os.path.join(TEST_FOLDER, "models", "invalid"+".json")
    with pytest.raises(ValueError) as excinfo:
        model = Model.load(filename)
    assert("invalid.json" in str(excinfo.value))

def test_json_invalid_include():
    """JSON exceptions should report file name, even for includes"""
    filename = os.path.join(TEST_FOLDER, "models", "invalid_include"+".json")
    with pytest.raises(ValueError) as excinfo:
        model = Model.load(filename)
    assert("invalid.json" in str(excinfo.value))


def test_variable_load():
    """Current timestep before model has started is undefined"""
    filename = os.path.join(TEST_FOLDER, "models", "demand_saving2_with_variables.json")
    model = Model.load(filename)

    # Test the correct number of each component is loaded
    assert len(model.variables) == 3
    assert len(model.objectives) == 1
    assert len(model.constraints) == 1
    # Test the names are as expected
    assert sorted([c.name for c in model.variables]) == ["demand_profile", "level1", "level2"]
    assert sorted([c.name for c in model.objectives]) == ["total_deficit", ]
    assert sorted([c.is_objective for c in model.objectives]) == ["minimise", ]
    assert sorted([c.name for c in model.constraints]) == ["min_volume", ]


def test_delta_greater_than_zero_days(simple_linear_model):
    """Test trying to use zero length timestep raises an error."""
    model = simple_linear_model
    model.timestepper.delta = 0

    with pytest.raises(ValueError):
        model.run()


def test_timestep_greater_than_zero_days():
    """Test trying to create zero length Timestep."""

    with pytest.raises(ValueError):
        # Test setting days <= 0 raises an error
        Timestep(pandas.Period('2019-01-01', freq='D'), 0, 0)


@pytest.mark.parametrize("freq, start_date, days_in_current_year, days_in_next_year", [
                                                                                       ("1D", "2012-12-31", 1, 0),
                                                                                       ("1D", "2011-12-31", 1, 0),
                                                                                       ("3D", "2012-12-30", 2, 1),
                                                                                       (7, "2012-12-29", 3, 4),
                                                                                       ("60h", "2012-12-31", 1, 1.5),
                                                                                       ("15h", "2012-12-31", 0.625, 0),
                                                                                       ("39h", "2012-12-31", 1, 0.625)
                                                                                      ])
def test_timestep_days_in_year_methods(simple_linear_model, freq, start_date, days_in_current_year, days_in_next_year):
    """Test the `days_in_current_year` and `days_in_next_year` methods of the timestep object"""
    simple_linear_model.timestepper.start = start_date
    simple_linear_model.timestepper.end = "2013-01-30"
    simple_linear_model.timestepper.delta = freq

    simple_linear_model.setup()
    simple_linear_model.step()

    ts = simple_linear_model.timestepper.current
    assert days_in_current_year == ts.days_in_current_year()
    assert days_in_next_year == ts.days_in_next_year()



""" Example Pywr model using a simple reservoir system.



"""
from pywr.model import Model
from pywr.recorders import TablesRecorder
import numpy as np
from matplotlib import pyplot as plt
import click
import time
import json
import pandas
MODEL_FILENAME = 'thames.json'

@click.group()
def cli():
    pass


@cli.command()
def run():

    # Run the model
    model = Model.load(MODEL_FILENAME)

    # Add a storage recorder
    TablesRecorder(model, 'thames_output.h5', parameters=[p for p in model.parameters])

    # Run the model
    stats = model.run()
    print(stats)
    stats_df = stats.to_dataframe()
    print(stats_df)

    keys_to_plot = (
        'time_taken_before',
        'solver_stats.bounds_update_nonstorage',
        'solver_stats.bounds_update_storage',
        'solver_stats.objective_update',
        'solver_stats.lp_solve',
        'solver_stats.result_update',
        'time_taken_after',
    )

    keys_to_tabulate = (
        'timesteps',
        'time_taken',
        'solver',
        'num_scenarios',
        'speed',
        'solver_name'
        'solver_stats.total',
        'solver_stats.number_of_rows',
        'solver_stats.number_of_cols',
        'solver_stats.number_of_nonzero',
        'solver_stats.number_of_routes',
        'solver_stats.number_of_nodes',
    )

    values = []
    labels = []
    explode = []
    solver_sub_total = 0.0
    for k in keys_to_plot:
        v = stats_df.loc[k][0]
        values.append(v)
        label = k.split('.', 1)[-1].replace('_', ' ').capitalize()
        explode.append(0.0)
        if k.startswith('solver_stats'):
            labels.append('Solver - {}'.format(label))
            solver_sub_total += v
        else:

            labels.append(label)

    values.append(stats_df.loc['solver_stats.total'][0] - solver_sub_total)
    labels.append('Solver - Other')
    explode.append(0.0)

    values.append(stats_df.loc['time_taken'][0] - sum(values))
    values = np.array(values) / sum(values)
    labels.append('Other')
    explode.append(0.0)

    fig, (ax1, ax2) = plt.subplots(figsize=(12, 4), ncols=2, sharey='row',
                                   gridspec_kw={'width_ratios': [2, 1]})

    print(values, labels)
    ax1.pie(values, explode=explode, labels=labels, autopct='%1.1f%%', startangle=90)
    ax1.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.

    cell_text = []
    for index, value in stats_df.iterrows():
        if index not in keys_to_tabulate:
            continue
        v = value[0]
        if isinstance(v, (float, np.float64, np.float32)):
            v = f'{v:.2f}'

        cell_text.append([index, v])

    tbl = ax2.table(cellText=cell_text, colLabels=['Statistic', 'Value'], loc='center')
    tbl.scale(1.5, 1.5)  # may help
    tbl.set_fontsize(14)
    ax2.axis('off')

    fig.savefig('run_statistics_w_tables.png', dpi=300)
    fig.savefig('run_statistics_w_tables.eps')

    plt.show()


@cli.command()
@click.option('--ext', default='png')
@click.option('--show/--no-show', default=False)
def figures(ext, show):

    for name, df in TablesRecorder.generate_dataframes('thames_output.h5'):
        df.columns = ['Very low', 'Low', 'Central', 'High', 'Very high']

        fig, (ax1, ax2) = plt.subplots(figsize=(12, 4), ncols=2, sharey='row',
                                       gridspec_kw={'width_ratios': [3, 1]})
        df['2100':'2125'].plot(ax=ax1)
        df.quantile(np.linspace(0, 1)).plot(ax=ax2)

        if name.startswith('reservoir'):
            ax1.set_ylabel('Volume [$Mm^3$]')
        else:
            ax1.set_ylabel('Flow [$Mm^3/day$]')

        for ax in (ax1, ax2):
            ax.set_title(name)
            ax.grid(True)
        plt.tight_layout()

        if ext is not None:
            fig.savefig(f'{name}.{ext}', dpi=300)

    if show:
        plt.show()


@cli.command('plot-res')
@click.option('--ext', default='png')
@click.option('--show/--no-show', default=False)
def plot_res(ext, show):

    end_year = '2105'

    data = {}
    for name, df in TablesRecorder.generate_dataframes('thames_output.h5'):
        df.columns = ['Very low', 'Low', 'Central', 'High', 'Very high']
        data[name] = df

    fig1, ax1 = plt.subplots(figsize=(16, 5), dpi=300)
    data['reservoir1'].loc[:end_year, 'Central'].plot(ax=ax1)
    ax1.set_ylabel('Volume [$Mm^3$]')
    plt.tight_layout()

    fig2, ax2 = plt.subplots(figsize=(16, 5), dpi=300)
    data['demand_saving_level'].loc[:end_year, 'Central'].plot(ax=ax2)
    ax2.set_ylabel('Demand saving level')
    plt.tight_layout()

    fig3, ax3 = plt.subplots(figsize=(16, 5), dpi=300)
    data['demand_max_flow'].loc[:end_year, 'Central'].plot(ax=ax3)
    ax3.set_ylabel('Demand [$Mm^3/day$]')
    plt.tight_layout()

    for ax in (ax1, ax2, ax3):
        ax.grid(True)

    if ext is not None:
        fig1.savefig(f'Reservoir.{ext}', dpi=300)
        fig2.savefig(f'Demand saving level.{ext}', dpi=300)
        fig3.savefig(f'Demand.{ext}', dpi=300)

    if show:
        plt.show()


@cli.command('plot-res2')
@click.option('--ext', default='png')
@click.option('--show/--no-show', default=False)
def plot_res2(ext, show):

    end_year = '2105'

    data = {}
    for name, df in TablesRecorder.generate_dataframes('thames_output.h5'):
        df.columns = ['Very low', 'Low', 'Central', 'High', 'Very high']
        data[name] = df

    fig1, ax1 = plt.subplots(figsize=(16, 5), dpi=300)
    data['reservoir1'].loc[:end_year].plot(ax=ax1)
    ax1.set_ylabel('Volume [$Mm^3$]')
    plt.legend()
    plt.tight_layout()

    fig2, ax2 = plt.subplots(figsize=(16, 5), dpi=300)
    data['reservoir1'].quantile(np.linspace(0, 1)).plot(ax=ax2)
    ax2.set_ylabel('Volume [$Mm^3$]')
    ax2.set_xlabel('Quantile')
    plt.tight_layout()

    fig3, ax3 = plt.subplots(figsize=(16, 5), dpi=300)
    df = data['demand_saving_level'].apply(pandas.Series.value_counts)
    df /= df.sum(axis=0)
    df.plot.bar(ax=ax3)
    ax3.set_ylabel('Proportion of time.')
    ax3.set_xlabel('Demand saving level')
    plt.tight_layout()

    for ax in (ax1, ax2, ax3):
        ax.grid(True)

    if ext is not None:
        fig1.savefig(f'Reservoir (scenarios).{ext}', dpi=300)
        fig2.savefig(f'Reservoir SDC (scenarios).{ext}', dpi=300)
        fig3.savefig(f'Demand saving level count (scenarios).{ext}', dpi=300)

    if show:
        plt.show()


@cli.command('plot-cc')
@click.option('--ext', default='png')
@click.option('--show/--no-show', default=False)
def plot_control_curves(ext, show):

    with open(MODEL_FILENAME) as fh:
        data = json.load(fh)

    parameters = data['parameters']

    fig, ax = plt.subplots(figsize=(8, 5), dpi=300)

    dates = pandas.date_range("2015-01-01", "2015-12-31")

    L1_values = np.array([parameters['level1']['values'][d.month-1] for d in dates])
    L2_values = np.array([parameters['level2']['values'][d.month-1] for d in dates])

    x = np.arange(0, len(dates)) + 1

    ax.fill_between(x, 1.0, L1_values, label='Level 0', alpha=0.8)
    ax.fill_between(x, L1_values, L2_values, label='Level 1', alpha=0.8)
    ax.fill_between(x, L2_values, 0.0, label='Level 2', alpha=0.8)

    plt.xlabel("Day of year")
    plt.ylabel("Reservoir volume [%]")

    plt.grid(True)
    plt.ylim([0.0, 1.0])
    plt.xlim(1, 365)
    plt.legend(["Level 0", "Level 1", "Level 2"], loc="upper right")
    ax.plot(x, L1_values, color='k', label=None)
    ax.plot(x, L2_values, color='k', label=None)
    plt.tight_layout()

    if ext is not None:
        fig.savefig(f'Control curve zones.{ext}', dpi=300)

    if show:
        plt.show()


@cli.command('plot-dsf')
@click.option('--ext', default='png')
@click.option('--show/--no-show', default=False)
def plot_demand_saving_factor(ext, show):

    with open(MODEL_FILENAME) as fh:
        data = json.load(fh)

    parameters = data['parameters']

    fig, ax = plt.subplots(figsize=(8, 5), dpi=300)

    dates = pandas.date_range("2015-01-01", "2015-12-31")

    L0_values = np.array([parameters['level0_factor']['values'] for d in dates])
    L1_values = np.array([parameters['level1_factor']['values'][d.month-1] for d in dates])
    L2_values = np.array([parameters['level2_factor']['values'][d.month-1] for d in dates])

    x = np.arange(0, len(dates)) + 1

    ax.plot(x, L0_values, label='Level 0')
    ax.plot(x, L1_values, label='Level 0')
    ax.plot(x, L2_values, label='Level 0')

    plt.xlabel("Day of year")
    plt.ylabel("Demand restriction factor")

    plt.grid(True)
    plt.ylim(0.6, 1.2)
    plt.xlim(1, 365)
    plt.legend(["Level 0", "Level 1", "Level 2"], loc="upper right")

    plt.tight_layout()

    if ext is not None:
        fig.savefig(f'Demand restriction factors.{ext}', dpi=300)

    if show:
        plt.show()


if __name__ == '__main__':
    cli()



import os
import numpy as np
cimport numpy as np
from scipy.interpolate import Rbf
import pandas
import json
import calendar
from libc.math cimport cos, M_PI
from libc.limits cimport INT_MIN, INT_MAX
from pywr.h5tools import H5Store
from pywr.hashes import check_hash
from .._core cimport is_leap_year
from ..dataframe_tools import align_and_resample_dataframe, load_dataframe, read_dataframe
import warnings


parameter_registry = {}


class UnutilisedDataWarning(Warning):
    """ Simple warning to indicate that not all data has been used. """
    pass

class TypeNotFoundError(KeyError):
    """
      Key Error, specifically designed for when the 'type' key is not found
      in a dataset. This takes the data value and outputs a summary of it, to
      aid in debugging.
    """
    def __init__(self, data):
        #Try to print out some sensible amount of data without overloading
        #the terminal with data. 1000 chars should be enough to get an idea
        #of what the data looks like. If more than 1000 chars, do a pandas-style
        #summary using ...
        data_str = json.dumps(data)
        if len(data_str) > 1000:
            data_summary = f"{data_str[:500]} ... {data_str[-500:]}"
        else:
            data_summary = data_str

        return f"Unable to find key 'type' in {data_summary}"

cdef class Parameter(Component):
    def __init__(self, *args, is_variable=False, **kwargs):
        super(Parameter, self).__init__(*args, **kwargs)
        self.is_variable = is_variable
        self.double_size = 0
        self.integer_size = 0
        self.is_constant = False

    @classmethod
    def register(cls):
        parameter_registry[cls.__name__.lower()] = cls

    @classmethod
    def unregister(cls):
        del(parameter_registry[cls.__name__.lower()])

    cpdef setup(self):
        super(Parameter, self).setup()
        cdef int num_comb
        if self.model.scenarios.combinations:
            num_comb = len(self.model.scenarios.combinations)
        else:
            num_comb = 1
        self.__values = np.empty([num_comb], np.float64)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        raise NotImplementedError("Parameter must be subclassed")

    cdef calc_values(self, Timestep timestep):
        # default implementation calls Parameter.value in loop
        cdef ScenarioIndex scenario_index
        cdef ScenarioCollection scenario_collection = self.model.scenarios
        for scenario_index in scenario_collection.combinations:
            self.__values[<int>(scenario_index.global_id)] = self.value(timestep, scenario_index)

    cpdef double get_value(self, ScenarioIndex scenario_index):
        return self.__values[<int>(scenario_index.global_id)]

    cpdef double[:] get_all_values(self):
        return self.__values

    cpdef double get_constant_value(self):
        """Return a constant value.
        
        This method should only be implemented and called if `is_constant` is True. 
        """
        raise NotImplementedError()

    cpdef set_double_variables(self, double[:] values):
        raise NotImplementedError()

    cpdef double[:] get_double_variables(self):
        raise NotImplementedError()

    cpdef double[:] get_double_lower_bounds(self):
        raise NotImplementedError()

    cpdef double[:] get_double_upper_bounds(self):
        raise NotImplementedError()

    cpdef set_integer_variables(self, int[:] values):
        raise NotImplementedError()

    cpdef int[:] get_integer_variables(self):
        raise NotImplementedError()

    cpdef int[:] get_integer_lower_bounds(self):
        raise NotImplementedError()

    cpdef int[:] get_integer_upper_bounds(self):
        raise NotImplementedError()

    property size:
        def __get__(self):
            warnings.warn("Use of the `size` property on Parameters has been deprecated."
                          "Please use either `double_size` or `integer_size` instead.", DeprecationWarning)
            return self.double_size

        def __set__(self, value):
            warnings.warn("Use of the `size` property on Parameters has been deprecated."
                          "Please use either `double_size` or `integer_size` instead.", DeprecationWarning)
            self.double_size = value

    @classmethod
    def load(cls, model, data):
        # If a scenario is given don't pass this to the load values methods
        scenario = data.pop('scenario', None)

        values = load_parameter_values(model, data)
        data.pop("values", None)
        data.pop("url", None)
        name = data.pop("name", None)
        comment = data.pop("comment", None)

        if scenario is not None:
            scenario = model.scenarios[scenario]
            # Only pass scenario object if one provided; most Parameter subclasses
            # do not accept a scenario argument.
            return cls(model, scenario=scenario, values=values, name=name, comment=None, **data)
        else:
            return cls(model, values=values, name=name, comment=None, **data)
Parameter.register()


cdef class ConstantParameter(Parameter):
    def __init__(self, model, value, lower_bounds=0.0, upper_bounds=np.inf, scale=1.0, offset=0.0, **kwargs):
        super(ConstantParameter, self).__init__(model, **kwargs)
        self._value = value
        self.scale = scale
        self.offset = offset
        self.double_size = 1
        self.integer_size = 0
        self.is_constant = True
        self._lower_bounds = np.ones(self.double_size) * lower_bounds
        self._upper_bounds = np.ones(self.double_size) * upper_bounds

    cdef calc_values(self, Timestep timestep):
        # constant parameter can just set the entire array to one value
        self.__values[...] = self.get_constant_value()

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self.get_constant_value()

    cpdef double get_constant_value(self):
        return self.offset + self._value * self.scale

    cpdef set_double_variables(self, double[:] values):
        self._value = values[0]

    cpdef double[:] get_double_variables(self):
        return np.array([self._value, ], dtype=np.float64)

    cpdef double[:] get_double_lower_bounds(self):
        return self._lower_bounds

    cpdef double[:] get_double_upper_bounds(self):
        return self._upper_bounds

    @classmethod
    def load(cls, model, data):
        if "value" in data:
            value = data.pop("value")
        else:
            value = load_parameter_values(model, data)
        parameter = cls(model, value, **data)
        return parameter

ConstantParameter.register()


cdef class DataFrameParameter(Parameter):
    """Timeseries parameter with automatic alignment and resampling

    Parameters
    ----------
    model : pywr.model.Model
    dataframe : pandas.DataFrame or pandas.Series
    scenario: pywr._core.Scenario (optional)
    """
    def __init__(self, model, dataframe, scenario=None, **kwargs):
        super(DataFrameParameter, self).__init__(model, *kwargs)
        self.dataframe = dataframe
        self.scenario = scenario

    cpdef setup(self):
        cdef Py_ssize_t i
        super(DataFrameParameter, self).setup()
        # align and resample the dataframe
        dataframe_resampled = align_and_resample_dataframe(self.dataframe, self.model.timestepper.datetime_index)
        if dataframe_resampled.ndim == 1:
            dataframe_resampled = pandas.DataFrame(dataframe_resampled)
        # dataframe should now have the correct number of timesteps for the model
        if len(dataframe_resampled) != len(self.model.timestepper):
            raise ValueError("Aligning DataFrame failed with a different length compared with model timesteps.")
        # check that if a 2D DataFrame is given that we also have a scenario assigned with it
        if dataframe_resampled.ndim == 2 and dataframe_resampled.shape[1] > 1:
            if self.scenario is None:
                raise ValueError("Scenario must be given for a DataFrame input with multiple columns.")
            if self.scenario.size != dataframe_resampled.shape[1]:
                raise ValueError("Scenario size ({}) is different to the number of columns ({}) "
                                 "in the DataFrame input.".format(self.scenario.size, dataframe_resampled.shape[1]))

        if self.scenario:
            # if possible, only load the data required
            scenario_indices = None
            # Default to index that is just out of bounds to cause IndexError if something goes wrong
            self._scenario_ids = np.ones(self.scenario.size, dtype=np.int32) * self.scenario.size

            # Calculate the scenario indices to load dependning on how scenario combinations are defined.
            if self.model.scenarios.user_combinations:
                scenario_indices = set()
                for user_combination in self.model.scenarios.user_combinations:
                    scenario_indices.add(user_combination[self._scenario_index])
                scenario_indices = sorted(list(scenario_indices))
            elif self.scenario.slice:
                scenario_indices = range(*self.scenario.slice.indices(self.scenario.slice.stop))
            else:
                # scenario is defined, but all data required
                self._scenario_ids = None

            if scenario_indices is not None:
                # Now load only the required data
                for n, i in enumerate(scenario_indices):
                    self._scenario_ids[i] = n
                dataframe_resampled = dataframe_resampled.iloc[:, scenario_indices]

        self._values = dataframe_resampled.values.astype(np.float64)
        if self.scenario is not None:
            self._scenario_index = self.model.scenarios.get_scenario_index(self.scenario)

    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        cdef double value
        cdef Py_ssize_t i = timestep.index
        cdef Py_ssize_t j

        if self.scenario is not None:
            j = scenario_index._indices[self._scenario_index]
            if self._scenario_ids is not None:
                j = self._scenario_ids[j]
            value = self._values[i, j]
        else:
            value = self._values[i, 0]
        return value

    @classmethod
    def load(cls, model, data):
        scenario = data.pop('scenario', None)
        if scenario is not None:
            scenario = model.scenarios[scenario]
        df = load_dataframe(model, data)
        return cls(model, df, scenario=scenario, **data)

DataFrameParameter.register()

cdef class ArrayIndexedParameter(Parameter):
    """Time varying parameter using an array and Timestep.index

    The values in this parameter are constant across all scenarios.
    """
    def __init__(self, model, values, *args, **kwargs):
        super(ArrayIndexedParameter, self).__init__(model, *args, **kwargs)
        self.values = np.asarray(values, dtype=np.float64)

    cdef calc_values(self, Timestep ts):
        # constant parameter can just set the entire array to one value
        self.__values[...] = self.values[ts.index]

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        """Returns the value of the parameter at a given timestep
        """
        return self.values[ts.index]
ArrayIndexedParameter.register()


cdef class ArrayIndexedScenarioParameter(Parameter):
    """A Scenario varying Parameter

    The values in this parameter are vary in time based on index and vary within a single Scenario.
    """
    def __init__(self, model, Scenario scenario, values, *args, **kwargs):
        """
        values should be an iterable that is the same length as scenario.size
        """
        super(ArrayIndexedScenarioParameter, self).__init__(model, *args, **kwargs)
        cdef int i
        values = np.asarray(values, dtype=np.float64)
        if values.ndim != 2:
            raise ValueError("Values must be two dimensional.")
        if scenario._size != values.shape[1]:
            raise ValueError("The size of the second dimension of values must equal the size of the scenario.")
        self.values = values
        self._scenario = scenario

    cpdef setup(self):
        super(ArrayIndexedScenarioParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        # This is a bit confusing.
        # scenario_indices contains the current scenario number for all
        # the Scenario objects in the model run. We have cached the
        # position of self._scenario in self._scenario_index to lookup the
        # correct number to use in this instance.
        return self.values[ts.index, scenario_index._indices[self._scenario_index]]

ArrayIndexedScenarioParameter.register()


cdef class TablesArrayParameter(IndexParameter):
    def __init__(self, model, h5file, node, where='/', scenario=None, **kwargs):
        """
        This Parameter reads array data from a PyTables HDF database.

        The parameter reads data using the PyTables array interface and therefore
        does not require loading the entire dataset in to memory. This is useful
        for large model runs.

        Parameters
        ----------
        h5file : tables.File or filename
            The tables file handle or filename to attach the CArray objects to. If a
            filename is given the object will open and close the file handles.
        node : string
            Name of the node in the tables database to read data from
        where : string
            Path to read the node from.
        scenario : Scenario
            Scenario to use as the second index in the array.
        """
        super(TablesArrayParameter, self).__init__(model, **kwargs)

        self.h5file = h5file
        self.h5store = None
        self.node = node
        self.where = where
        self.scenario = scenario

        # Private attributes, initialised during setup()
        self._values_dbl = None  # Stores the loaded data if float
        self._values_int = None  # Stores the loaded data if integer
        # If a scenario is present this is the index in the model list of scenarios
        self._scenario_index = -1
        self._scenario_ids = None  # Lookup of scenario index to the loaded data index

    cpdef setup(self):
        cdef Py_ssize_t n, i

        super(TablesArrayParameter, self).setup()
        self._scenario_index = -1
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        if self.scenario is not None:
            self._scenario_index = self.model.scenarios.get_scenario_index(self.scenario)

        self.h5store = H5Store(self.h5file, None, "r")
        node = self.h5store.file.get_node(self.where, self.node)
        if not node.dtype in (np.float32, np.float64, np.int8, np.int16, np.int32):
            raise TypeError("Unexpected dtype in array: {}".format(node.dtype))

        # check the shape of the data is valid
        if self.scenario is not None:
            if node.shape[1] < self.scenario.size:
                raise IndexError('The length of the second dimension ({:d}) of the tables node ({}:{}) '
                                 'should be the same as the size of the specified Scenario ({:d}).'
                                 .format(node.shape[1], node._v_file.filename, node._v_pathname, self.scenario.size))
            elif node.shape[1] > self.scenario.size:
                warnings.warn('The length of the second dimension ({:d}) of the tables node ({}:{}) '
                              'is greater than the size of the specified Scenario ({:d}). '
                              'Not all data is being used!'.format(node.shape[1], node._v_file.filename, node._v_pathname, self.scenario.size),
                              UnutilisedDataWarning)
        if node.shape[0] < len(self.model.timestepper):
            raise IndexError('The length of the first dimension ({:d}) of the tables node ({}:{}) '
                             'should be equal to or greater than the number of timesteps.'
                             .format(node.shape[0], node._v_file.filename, node._v_pathname, len(self.model.timestepper)))
        elif node.shape[0] > len(self.model.timestepper):
            warnings.warn('The length of the first dimension ({:d}) of the tables node ({}:{}) '
                          'is greater than the number of timesteps. Not all data is being used!'
                          .format(node.shape[0], node._v_file.filename, node._v_pathname, len(self.model.timestepper)),
                          UnutilisedDataWarning)

        # detect data type and read into memoryview
        self._values_dbl = None
        self._values_int = None
        if self.scenario:
            # if possible, only load the data required
            scenario_indices = None
            # Default to index that is just out of bounds to cause IndexError if something goes wrong
            self._scenario_ids = np.ones(self.scenario.size, dtype=np.int32) * self.scenario.size

            # Calculate the scenario indices to load dependning on how scenario combinations are defined.
            if self.model.scenarios.user_combinations:
                scenario_indices = set()
                for user_combination in self.model.scenarios.user_combinations:
                    scenario_indices.add(user_combination[self._scenario_index])
                scenario_indices = sorted(list(scenario_indices))
            elif self.scenario.slice:
                scenario_indices = range(*self.scenario.slice.indices(self.scenario.slice.stop))
            else:
                # scenario is defined, but all data required
                self._scenario_ids = None

            if scenario_indices is not None:
                # Now load only the required data
                for n, i in enumerate(scenario_indices):
                    self._scenario_ids[i] = n

                if node.dtype in (np.float32, np.float64):
                    self._values_dbl = node[:len(self.model.timestepper), scenario_indices].astype(np.float64)
                else:
                    self._values_int = node[:len(self.model.timestepper), scenario_indices].astype(np.int32)

        if node.dtype in (np.float32, np.float64):
            if self._values_dbl is None:
                self._values_dbl = node.read().astype(np.float64)
            # negative values are often erroneous
            if np.min(self._values_dbl) < 0.0:
                warnings.warn('Negative values in input file "{}" from node: {}'.format(self.h5file, self.node))
            if not np.all(np.isfinite(self._values_dbl)):
                raise ValueError('Non-finite values in input file "{}" from node: {}'.format(self.h5file, self.node))
        else:
            if self._values_int is None:
                self._values_int = node.read().astype(np.int32)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef Py_ssize_t i = ts.index
        cdef Py_ssize_t j
        if self._values_dbl is None:
            return float(self.index(ts, scenario_index))
        # Support 1D and 2D indexing when scenario is or is not given.
        if self._scenario_index == -1:
            return self._values_dbl[i, 0]
        else:
            j = scenario_index._indices[self._scenario_index]
            if self._scenario_ids is not None:
                j = self._scenario_ids[j]
            return self._values_dbl[i, j]

    cpdef int index(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef Py_ssize_t i = ts.index
        cdef Py_ssize_t j
        if self._values_int is None:
            return int(self.value(ts, scenario_index))
        # Support 1D and 2D indexing when scenario is or is not given.
        if self._scenario_index == -1:
            return self._values_int[i, 0]
        else:
            j = scenario_index._indices[self._scenario_index]
            if self._scenario_ids is not None:
                j = self._scenario_ids[j]
            return self._values_int[i, j]

    cpdef finish(self):
        self.h5store = None

    @classmethod
    def load(cls, model, data):
        scenario = data.pop('scenario', None)
        if scenario is not None:
            scenario = model.scenarios[scenario]

        url = data.pop('url')
        if not os.path.isabs(url) and model.path is not None:
            url = os.path.join(model.path, url)
        node = data.pop('node')
        where = data.pop('where', '/')

        # Check hashes if given before reading the data
        checksums = data.pop('checksum', {})
        for algo, hash in checksums.items():
            check_hash(url, hash, algorithm=algo)

        return cls(model, url, node, where=where, scenario=scenario)
TablesArrayParameter.register()


cdef class ConstantScenarioParameter(Parameter):
    """A Scenario varying Parameter

    The values in this parameter are constant in time, but vary within a single Scenario.
    """
    def __init__(self, model, Scenario scenario, values, *args, **kwargs):
        """
        values should be an iterable that is the same length as scenario.size
        """
        super(ConstantScenarioParameter, self).__init__(model, *args, **kwargs)
        cdef int i
        if scenario._size != len(values):
            raise ValueError("The number of values must equal the size of the scenario.")
        self._values = np.empty(scenario._size)
        for i in range(scenario._size):
            self._values[i] = values[i]
        self._scenario = scenario

    cpdef setup(self):
        super(ConstantScenarioParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        # This is a bit confusing.
        # scenario_indices contains the current scenario number for all
        # the Scenario objects in the model run. We have cached the
        # position of self._scenario in self._scenario_index to lookup the
        # correct number to use in this instance.
        return self._values[scenario_index._indices[self._scenario_index]]
ConstantScenarioParameter.register()


cdef class ArrayIndexedScenarioMonthlyFactorsParameter(Parameter):
    """Time varying parameter using an array and Timestep.index with
    multiplicative factors per Scenario
    """
    def __init__(self, model, Scenario scenario, values, factors, *args, **kwargs):
        """
        values is the baseline timeseries data that is perturbed by a factor. The
        factor is taken from factors which is shape (scenario.size, 12). Therefore
        factors vary with the individual scenarios in scenario and month.
        """
        super(ArrayIndexedScenarioMonthlyFactorsParameter, self).__init__(model, *args, **kwargs)

        values = np.asarray(values, dtype=np.float64)
        factors = np.asarray(factors, dtype=np.float64)
        if factors.ndim != 2:
            raise ValueError("Factors must be two dimensional.")

        if factors.shape[0] != scenario._size:
            raise ValueError("First dimension of factors must be the same size as scenario.")
        if factors.shape[1] != 12:
            raise ValueError("Second dimension of factors must be 12.")
        self._scenario = scenario
        self._values = values
        self._factors = factors

    cpdef setup(self):
        super(ArrayIndexedScenarioMonthlyFactorsParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        # This is a bit confusing.
        # scenario_indices contains the current scenario number for all
        # the Scenario objects in the model run. We have cached the
        # position of self._scenario in self._scenario_index to lookup the
        # correct number to use in this instance.
        cdef int imth = ts.month-1
        cdef int i = scenario_index._indices[self._scenario_index]
        return self._values[ts.index]*self._factors[i, imth]

    @classmethod
    def load(cls, model, data):
        scenario = data.pop("scenario", None)
        if scenario is not None:
            scenario = model.scenarios[scenario]

        if isinstance(data["values"], list):
            values = np.asarray(data["values"], np.float64)
        elif isinstance(data["values"], dict):
            values = load_parameter_values(model, data["values"])
        else:
            raise TypeError("Unexpected type for \"values\" in {}".format(cls.__name__))

        if isinstance(data["factors"], list):
            factors = np.asarray(data["factors"], np.float64)
        elif isinstance(data["factors"], dict):
            factors = load_parameter_values(model, data["factors"])
        else:
            raise TypeError("Unexpected type for \"factors\" in {}".format(cls.__name__))

        return cls(model, scenario, values, factors)

ArrayIndexedScenarioMonthlyFactorsParameter.register()


cdef class DailyProfileParameter(Parameter):
    """ An annual profile consisting of daily values.

    This parameter provides a repeating annual profile with a daily resolution. A total of 366 values
    must be provided. These values are coerced to a `numpy.array` internally.

    Parameters
    ----------
    values : iterable, array
        The 366 values that represent the daily profile.

    """
    def __init__(self, model, values, *args, **kwargs):
        super(DailyProfileParameter, self).__init__(model, *args, **kwargs)
        v = np.squeeze(np.array(values))
        if v.ndim != 1:
            raise ValueError("values must be 1-dimensional.")
        if len(values) != 366:
            raise ValueError("366 values must be given for a daily profile.")
        self._values = v

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self._values[ts.dayofyear_index]
DailyProfileParameter.register()

cdef class WeeklyProfileParameter(Parameter):
    """Weekly profile (52-week year)

    The last week of the year will have more than 7 days, as 365 / 7 is not whole.
    """
    def __init__(self, model, values, *args, **kwargs):
        super(WeeklyProfileParameter, self).__init__(model, *args, **kwargs)
        v = np.squeeze(np.array(values))
        if v.ndim != 1:
            raise ValueError("values must be 1-dimensional.")
        if len(values) == 53:
            values = values[:52]
            warnings.warn("Truncating 53 week profile to 52 weeks.")
        if len(values) != 52:
            raise ValueError("52 values must be given for a weekly profile.")
        self._values = v

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self._values[ts.week_index]
WeeklyProfileParameter.register()


cdef class MonthlyProfileParameter(Parameter):
    """Parameter which provides a monthly profile.

    The monthly profile returns a different value based on the month of the current
    time-step. By default this creates a piecewise profile with a step change at the
    beginning of each month. An optional `interp_day` keyword can instead create a
    linearly interpolated daily profile assuming the given values correspond to either
    the first or last day of the month.

    Parameters
    ----------
    values : iterable, array
        The 12 values that represent the monthly profile.
    lower_bounds : float or array_like (default=0.0)
        Defines the lower bounds when using optimisation. If a float given, same bound applied for every
        month. Otherwise an array like object of length 12 should be given for as separate value each month.
    upper_bounds : float or array_like (default=np.inf)
        Defines the upper bounds when using optimisation. If a float given, same bound applied for every
        month. Otherwise an array like object of length 12 should be given for as separate value each month.
    interp_day : str or None (default=None)
        If `interp_day` is None then no interpolation is undertaken, and the parameter
         returns values representing a piecewise monthly profile. Otherwise `interp_day`
         must be a string of either "first" or "last" representing which day of the month
         each of the 12 values represents. The parameter then returns linearly
         interpolated values between the given day of the month.


    See also
    --------
    ScenarioMonthlyProfileParameter
    ArrayIndexedScenarioMonthlyFactorsParameter
    """
    def __init__(self, model, values, lower_bounds=0.0, upper_bounds=np.inf, interp_day=None, **kwargs):
        super(MonthlyProfileParameter, self).__init__(model, **kwargs)
        self.double_size = 12
        self.integer_size = 0
        if len(values) != self.double_size:
            raise ValueError("12 values must be given for a monthly profile.")
        self._values = np.array(values)
        self.interp_day = interp_day

        if np.isscalar(lower_bounds):
            lb = np.ones(self.double_size) * lower_bounds
        else:
            lb = np.array(lower_bounds)
            if len(lb) != self.double_size:
                raise ValueError("Lower bounds must be a scalar or array like of 12 values.")
        self._lower_bounds = lb

        if np.isscalar(upper_bounds):
            ub = np.ones(self.double_size) * upper_bounds
        else:
            ub = np.array(upper_bounds)
            if len(ub) != self.double_size:
                raise ValueError("Upper bounds must be a scalar or array like of 12 values.")
        self._upper_bounds = ub


    cpdef reset(self):
        Parameter.reset(self)
        # The interpolated profile is recalculated during reset so that
        # it will update when the _values array is updated via `set_double_variables`
        # and the model is rerun. I.e. during optimisation (where setup is not redone).
        if self.interp_day is not None:
            self._interpolate()

    cpdef _interpolate(self):

        # Create an array to save the daily profile in.
        self._interp_values = np.zeros(366)
        cdef int i = 0
        cdef int mth

        # Create interpolation knots depending on values
        if self.interp_day == 'first':
            x = [1]  # First month
            y = []
            for mth in range(1, 13):
                x.append(x[-1] + calendar.monthrange(2015, mth)[1])
                y.append(self._values[mth-1])
            y.append(self._values[0])
        elif self.interp_day == 'last':
            x = [0]  # End of previous year
            y = [self._values[11]]  # Use value from December
            for mth in range(1, 13):
                x.append(x[-1] + calendar.monthrange(2015, mth)[1])
                y.append(self._values[mth-1])
        else:
            raise ValueError(f'Interpolation day "{self.interp_day}" not supported.')

        # Do the interpolation
        values = np.interp(np.arange(365) + 1, x, y)
        # Make the daily profile of 366 values repeating the same value for 28th & 29th Feb.
        for i in range(365):
            if i < 58:
                self._interp_values[i] = values[i]
            elif i == 58:
                self._interp_values[i] = values[i]
                self._interp_values[i+1] = values[i]
            elif i > 58:
                self._interp_values[i+1] = values[i]

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        if self.interp_day is None:
            return self._values[ts.month-1]
        else:
            return self._interp_values[ts.dayofyear_index]

    cpdef set_double_variables(self, double[:] values):
        self._values[...] = values

    cpdef double[:] get_double_variables(self):
        # Make sure we return a copy of the data instead of a view.
        return np.array(self._values).copy()

    cpdef double[:] get_double_lower_bounds(self):
        return self._lower_bounds

    cpdef double[:] get_double_upper_bounds(self):
        return self._upper_bounds
MonthlyProfileParameter.register()


cdef class ScenarioMonthlyProfileParameter(Parameter):
    """ Parameter that provides a monthly profile per scenario

    Behaviour is the same as `MonthlyProfileParameter` except a different
    profile is returned for each ensemble in a given scenario.

    See also
    --------
    MonthlyProfileParameter
    ArrayIndexedScenarioMonthlyFactorsParameter
    """
    def __init__(self, model, Scenario scenario, values, **kwargs):
        super(ScenarioMonthlyProfileParameter, self).__init__(model, **kwargs)

        if values.ndim != 2:
            raise ValueError("Factors must be two dimensional.")

        if scenario._size != values.shape[0]:
            raise ValueError("First dimension of factors must be the same size as scenario.")
        if values.shape[1] != 12:
            raise ValueError("Second dimension of factors must be 12.")
        self._scenario = scenario
        self._values = np.array(values)

    cpdef setup(self):
        super(ScenarioMonthlyProfileParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self._values[scenario_index._indices[self._scenario_index], ts.month-1]

ScenarioMonthlyProfileParameter.register()

cdef class ScenarioWeeklyProfileParameter(Parameter):
    """Parameter that provides a weekly profile per scenario

    This parameter provides a repeating annual profile with a weekly resolution. A
    different profile is returned for each member of a given scenario

    Parameters
    ----------
    scenario: Scenario
        Scenario object over which different profiles should be provided.
    values : iterable, array
        Length of 1st dimension should equal the number of members in the scenario object
        and the length of the second dimension should be 52

    """
    def __init__(self, model, Scenario scenario, values, *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        values = np.array(values)
        if values.ndim != 2:
            raise ValueError("Factors must be two dimensional.")
        if scenario._size != values.shape[0]:
            raise ValueError("First dimension of factors must be the same size as scenario.")
        if values.shape[1] != 52:
            raise ValueError("52 values must be given for a weekly profile.")
        self._values = values
        self._scenario = scenario

    cpdef setup(self):
        super(ScenarioWeeklyProfileParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self._values[scenario_index._indices[self._scenario_index], ts.week_index]

ScenarioWeeklyProfileParameter.register()

cdef class ScenarioDailyProfileParameter(Parameter):
    """Parameter which provides a daily profile per scenario.

    This parameter provides a repeating annual profile with a daily resolution. A
    different profile is returned for each member of a given scenario

    Parameters
    ----------
    scenario: Scenario
        Scenario object over which different profiles should be provided
    values : iterable, array
        Length of 1st dimension should equal the number of members in the scenario object
        and the length of the second dimension should be 366

    """
    def __init__(self, model, Scenario scenario, values, *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        values = np.array(values)
        if values.ndim != 2:
            raise ValueError("Factors must be two dimensional.")
        if scenario._size != values.shape[0]:
            raise ValueError("First dimension of factors must be the same size as scenario.")
        if values.shape[1] != 366:
            raise ValueError("366 values must be given for a daily profile.")
        self._values = values
        self._scenario = scenario

    cpdef setup(self):
        super(ScenarioDailyProfileParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int i = ts.dayofyear_index
        return self._values[scenario_index._indices[self._scenario_index], i]

ScenarioDailyProfileParameter.register()


cdef class UniformDrawdownProfileParameter(Parameter):
    """Parameter which provides a uniformly reducing value from one to zero.

     This parameter is intended to be used with an `AnnualVirtualStorage` node to provide a profile
     that represents perfect average utilisation of the annual volume. It returns a value of 1 on the
     reset day, and subsequently reduces by 1/366 every day afterward.

    Parameters
    ----------
    reset_day: int
        The day of the month (1-31) to reset the volume to the initial value.
    reset_month: int
        The month of the year (1-12) to reset the volume to the initial value.

    See also
    --------
    AnnualVirtualStorage
    """
    def __init__(self, model, reset_day=1, reset_month=1, **kwargs):
        super().__init__(model, **kwargs)
        self.reset_day = reset_day
        self.reset_month = reset_month

    cpdef reset(self):
        super(UniformDrawdownProfileParameter, self).reset()
        # Reset day of the year based on a leap year.
        # Note that this is zero-based
        self._reset_idoy = pandas.Period(year=2016, month=self.reset_month, day=self.reset_day, freq='D').dayofyear - 1

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int current_idoy = ts.dayofyear_index
        cdef int total_days_in_period
        cdef int days_into_period
        cdef int year = ts.year

        days_into_period = current_idoy - self._reset_idoy
        if days_into_period < 0:
            # We're not past the reset day yet; use the previous year
            year -= 1

        if self._reset_idoy > 59:
            # Reset occurs after February therefore next year's February might be a leap year?
            year += 1

        # Determine the number of days in the period based on whether there is a leap year or not in the current period
        if is_leap_year(year):
            total_days_in_period = 366
        else:
            total_days_in_period = 365

        # Now determine number of days we're into the period if it has wrapped around to a new year
        if days_into_period < 0:
            days_into_period += 366
            # Need to adjust for post 29th Feb in non-leap years.
            # Recall `current_idoy` was incremented by 1 if it is a non-leap already (hence comparison to 59)
            if not is_leap_year(ts.year) and current_idoy > 59:
                days_into_period -= 1

        return 1.0 - days_into_period / total_days_in_period

    @classmethod
    def load(cls, model, data):
        return cls(model, **data)
UniformDrawdownProfileParameter.register()


cdef class RbfProfileParameter(Parameter):
    """Parameter which interpolates a daily profile using a radial basis function (RBF).

    The daily profile is computed during model `reset` using a radial basis function with
    day-of-year as the independent variables. The days of the year are defined by the user
    alongside the values to use on each of those days for the interpolation. The first
    day of the years should always be one, and its value is repeated as the 366th value.
    In addition the second and penultimate values are mirrored to encourage a consistent
    gradient to appear across the boundary. The RBF calculations are undertaken using
    the `scipy.interpolate.Rbf` object, please refer to Scipy's documentation for more
    information.

    Parameters
    ----------
    days_of_year : iterable, integer
        The days of the year at which the interpolation values are defined. The first
        value should be one.
    values : iterable, float
        Values to use for interpolation corresponding to the `days_of_year`.
    lower_bounds : float or array_like (default=0.0)
        Defines the lower bounds when using optimisation. If a float given, same bound applied for every day of the
        year. Otherwise an array like object of length equal to the number of days of the year should be given.
    upper_bounds : float or array_like (default=np.inf)
        Defines the upper bounds when using optimisation. If a float given, same bound applied for every day of the
        year. Otherwise an array like object of length equal to the number of days of the year should be given.
    variable_days_of_year_range : int (default=0)
        The maximum bounds (positive or negative) for the days of year during optimisation. A non-zero value
        will cause the days of the year values to be exposed as integer variables (except the first value which
        remains at day 1). This value is bounds on those variables as maximum shift from the given `days_of_year`.
    min_value, max_value : float
        Optionally cap the interpolated daily profile to a minimum and/or maximum value. The default values
        are negative and positive infinity for minimum and maximum respectively.
    rbf_kwargs: Optional, dict
        Optional dictionary of keyword arguments to base to the Rbf object.

    """
    def __init__(self, model, days_of_year, values, lower_bounds=0.0, upper_bounds=np.inf, rbf_kwargs=None,
                 variable_days_of_year_range=0, min_value=-np.inf, max_value=np.inf, **kwargs):
        super(RbfProfileParameter, self).__init__(model, **kwargs)

        if len(days_of_year) != len(values):
            raise ValueError(f"The length of values ({len(values)}) must equal the length of "
                             f"`days_of_year` ({len(days_of_year)}).")

        self.variable_days_of_year_range = variable_days_of_year_range
        self.double_size = len(values)
        self._values = np.array(values, dtype=np.float64)
        self.days_of_year = days_of_year
        self.min_value = min_value
        self.max_value = max_value

        if np.isscalar(lower_bounds):
            lb = np.ones(self.double_size) * lower_bounds
        else:
            lb = np.array(lower_bounds)
            if len(lb) != self.double_size:
                raise ValueError("Lower bounds must be a scalar or array like with length equivalent to rbf values")
        self._lower_bounds = lb

        if np.isscalar(upper_bounds):
            ub = np.ones(self.double_size) * upper_bounds
        else:
            ub = np.array(upper_bounds)
            if len(ub) != self.double_size:
               raise ValueError("Upper bounds must be a scalar or array like with length equivalent to rbf values")
        self._upper_bounds = ub

        if self.variable_days_of_year_range > 0:
            if np.any(np.diff(self.days_of_year) <= 2*self.variable_days_of_year_range):
                raise ValueError(f"The days of the year are too close together for the given "
                                 f"`variable_days_of_year_range`. This could cause the optimised days"
                                 f"of the year to overlap and become out of order.  Either increase the"
                                 f"spacing of the days of the year or reduce `variable_days_of_year_range` to"
                                 f"less than half the closest distance between the days of the year.")
            self.integer_size = len(values) - 1
            self._doy_lower_bounds = np.array([d - self.variable_days_of_year_range
                                               for d in self.days_of_year[1:]], dtype=np.int32)
            self._doy_upper_bounds = np.array([d + self.variable_days_of_year_range
                                               for d in self.days_of_year[1:]], dtype=np.int32)
        else:
            self.integer_size = 0

        self.rbf_kwargs = rbf_kwargs if rbf_kwargs is not None else {}

    property days_of_year:
        def __get__(self):
            return np.array(self._days_of_year)
        def __set__(self, values):
            values = np.array(values, dtype=np.int32)
            if values[0] != 1:
                raise ValueError('The first day of the years must be 1.')
            if len(values) < 3:
                raise ValueError('At least 3 days of the year are required.')
            if np.any(np.diff(values) <= 0):
                raise ValueError('The days of the year should be strictly monotonically increasing.')
            if np.any(0 > values > 365):
                raise ValueError('Days of the years should be between 1 and 365 inclusive.')
            self._days_of_year = values

    cpdef reset(self):
        Parameter.reset(self)
        # The interpolated profile is recalculated during reset so that
        # it will update when the _values array is updated via `set_double_variables`
        # and the model is rerun. I.e. during optimisation (where setup is not redone).
        self._interpolate()

    cpdef _interpolate(self):
        cdef int i
        cdef double[:] values
        cdef double v

        days_of_year = list(self._days_of_year)
        # Append day 365 to the list and mirror the penultimate and second DOY at the start and end
        # of the list respectively. This helps ensure the gradient is roughly the same across the boundary
        # between days 365 and 0.
        days_of_year = [days_of_year[-1]-365] + list(days_of_year) + [366, 366+days_of_year[1]-1]
        # Create the corresponding y values including the mirrored entries
        y = list(self._values)
        y = [y[-1]] + y + [y[0], y[1]]
        rbfi = Rbf(days_of_year, y, **self.rbf_kwargs)

        # Do the interpolation
        values = rbfi(np.arange(365) + 1)

        # Create an array to save the daily profile in.
        self._interp_values = np.zeros(366)
        # Make the daily profile of 366 values repeating the same value for 28th & 29th Feb.
        for i in range(365):
            v = max(min(values[i], self.max_value), self.min_value)
            if i < 58:
                self._interp_values[i] = v
            elif i == 58:
                self._interp_values[i] = v
                self._interp_values[i+1] = v
            elif i > 58:
                self._interp_values[i+1] = v

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int i = ts.dayofyear_index
        return self._interp_values[i]

    # Double variables are for the known interpolation values (y-axis)
    cpdef set_double_variables(self, double[:] values):
        self._values[...] = values

    cpdef double[:] get_double_variables(self):
        # Make sure we return a copy of the data instead of a view.
        return np.array(self._values).copy()

    cpdef double[:] get_double_lower_bounds(self):
        return self._lower_bounds

    cpdef double[:] get_double_upper_bounds(self):
        return self._upper_bounds

    # Integer variables are for the days of the year positions (if optimised)
    cpdef set_integer_variables(self, int[:] values):
        self.days_of_year = [1] + np.array(values).tolist()

    cpdef int[:] get_integer_variables(self):
        return np.array(self.days_of_year[1:], dtype=np.int32)

    cpdef int[:] get_integer_lower_bounds(self):
        return self._doy_lower_bounds

    cpdef int[:] get_integer_upper_bounds(self):
        return self._doy_upper_bounds

    @classmethod
    def load(cls, model, data):
        return cls(model, **data)
RbfProfileParameter.register()


cdef class IndexParameter(Parameter):
    """Base parameter providing an `index` method

    See also
    --------
    IndexedArrayParameter
    ControlCurveIndexParameter
    """
    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        """Returns the current index as a float"""
        # return index as a float
        return float(self.get_index(scenario_index))

    cpdef int index(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        """Returns the current index"""
        # return index as an integer
        return 0

    cpdef setup(self):
        super(IndexParameter, self).setup()
        cdef int num_comb
        if self.model.scenarios.combinations:
            num_comb = len(self.model.scenarios.combinations)
        else:
            num_comb = 1
        self.__indices = np.empty([num_comb], np.int32)

    cdef calc_values(self, Timestep timestep):
        cdef ScenarioIndex scenario_index
        cdef ScenarioCollection scenario_collection = self.model.scenarios
        for scenario_index in scenario_collection.combinations:
            self.__indices[<int>(scenario_index.global_id)] = self.index(timestep, scenario_index)
            self.__values[<int>(scenario_index.global_id)] = self.value(timestep, scenario_index)

    cpdef int get_index(self, ScenarioIndex scenario_index):
        return self.__indices[<int>(scenario_index.global_id)]

    cpdef int[:] get_all_indices(self):
        return self.__indices
IndexParameter.register()


cdef class ConstantScenarioIndexParameter(IndexParameter):
    """A Scenario varying IndexParameter

    The values in this parameter are constant in time, but vary within a single Scenario.
    """
    def __init__(self, model, Scenario scenario, values, *args, **kwargs):
        """
        values should be an iterable that is the same length as scenario.size
        """
        super(ConstantScenarioIndexParameter, self).__init__(model, *args, **kwargs)
        cdef int i
        if scenario._size != len(values):
            raise ValueError("The number of values must equal the size of the scenario.")
        self._values = np.empty(scenario._size, dtype=np.int32)
        for i in range(scenario._size):
            self._values[i] = values[i]
        self._scenario = scenario

    cpdef setup(self):
        super(ConstantScenarioIndexParameter, self).setup()
        # This setup must find out the index of self._scenario in the model
        # so that it can return the correct value in value()
        self._scenario_index = self.model.scenarios.get_scenario_index(self._scenario)

    cpdef int index(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        # This is a bit confusing.
        # scenario_indices contains the current scenario number for all
        # the Scenario objects in the model run. We have cached the
        # position of self._scenario in self._scenario_index to lookup the
        # correct number to use in this instance.
        return self._values[scenario_index._indices[self._scenario_index]]
ConstantScenarioIndexParameter.register()


cdef class IndexedArrayParameter(Parameter):
    """Parameter which uses an IndexParameter to index an array of Parameters

    An example use of this parameter is to return a demand saving factor (as
    a float) based on the current demand saving level (calculated by an
    `IndexParameter`).

    Parameters
    ----------
    index_parameter : `IndexParameter`
    params : iterable of `Parameters` or floats


    Notes
    -----
    Float arguments `params` are converted to `ConstantParameter`
    """
    def __init__(self, model, index_parameter, params, **kwargs):
        super(IndexedArrayParameter, self).__init__(model, **kwargs)
        assert(isinstance(index_parameter, IndexParameter))
        self.index_parameter = index_parameter
        self.children.add(index_parameter)

        self.params = []
        for p in params:
            if not isinstance(p, Parameter):
                from pywr.parameters import ConstantParameter
                p = ConstantParameter(model, p)
            self.params.append(p)

        for param in self.params:
            self.children.add(param)
        self.children.add(index_parameter)

    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        """Returns the value of the Parameter at the current index"""
        cdef int index
        index = self.index_parameter.get_index(scenario_index)
        cdef Parameter parameter = self.params[index]
        return parameter.get_value(scenario_index)

    @classmethod
    def load(cls, model, data):
        index_parameter = load_parameter(model, data.pop("index_parameter"))
        try:
            parameters = data.pop("params")
        except KeyError:
            parameters = data.pop("parameters")
        parameters = [load_parameter(model, parameter_data) for parameter_data in parameters]
        return cls(model, index_parameter, parameters, **data)
IndexedArrayParameter.register()


cdef class AnnualHarmonicSeriesParameter(Parameter):
    """ A `Parameter` which returns the value from an annual harmonic series

    This `Parameter` comprises a series N cosine function with a period of 365
     days. The calculation is performed using the Julien day of the year minus 1
     This causes a small discontinuity in non-leap years.

    .. math:: f(t) = A + \sum_{n=1}^N A_n\cdot \cos((2\pi nt)/365+\phi_n)

    Parameters
    ----------

    mean : float
        Mean value for the series (i.e. the position of zeroth harmonic)
    amplitudes : array_like
        The amplitudes for the N harmonic cosine functions. Must be the same
        length as phases.
    phases : array_like
        The phase shift of the N harmonic cosine functions. Must be the same
        length as amplitudes.

    """
    def __init__(self, model, mean, amplitudes, phases, *args, **kwargs):
        if len(amplitudes) != len(phases):
            raise ValueError("The number  of amplitudes and phases must be the same.")
        n = len(amplitudes)
        self.mean = mean
        self._amplitudes = np.array(amplitudes)
        self._phases = np.array(phases)

        self._mean_lower_bounds = kwargs.pop('mean_lower_bounds', 0.0)
        self._mean_upper_bounds = kwargs.pop('mean_upper_bounds', np.inf)
        self._amplitude_lower_bounds = np.ones(n)*kwargs.pop('amplitude_lower_bounds', 0.0)
        self._amplitude_upper_bounds = np.ones(n)*kwargs.pop('amplitude_upper_bounds', np.inf)
        self._phase_lower_bounds = np.ones(n)*kwargs.pop('phase_lower_bounds', 0.0)
        self._phase_upper_bounds = np.ones(n)*kwargs.pop('phase_upper_bounds', np.pi*2)
        super(AnnualHarmonicSeriesParameter, self).__init__(model, *args, **kwargs)
        # Size must be set after call to super.
        self.double_size = 1 + 2*n
        self._value_cache = 0.0
        self._ts_index_cache = -1

    @classmethod
    def load(cls, model, data):
        mean = data.pop('mean')
        amplitudes = data.pop('amplitudes')
        phases = data.pop('phases')

        return cls(model, mean, amplitudes, phases, **data)

    property amplitudes:
        def __get__(self):
            return np.asarray(self._amplitudes)

    property phases:
        def __get__(self):
            return np.asarray(self._phases)

    cpdef reset(self):
        Parameter.reset(self)
        self._value_cache = 0.0
        self._ts_index_cache = -1

    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        cdef int ts_index = timestep.index
        cdef int doy = timestep.dayofyear - 1
        cdef int n = self._amplitudes.shape[0]
        cdef int i
        cdef double val
        if ts_index == self._ts_index_cache:
            val = self._value_cache
        else:
            val = self.mean
            for i in range(n):
                val += self._amplitudes[i]*cos(doy*(i+1)*M_PI*2/365 + self._phases[i])
            self._value_cache = val
            self._ts_index_cache = ts_index
        return val

    cpdef set_double_variables(self, double[:] values):
        n = len(self.amplitudes)
        self.mean = values[0]
        self._amplitudes[...] = values[1:n+1]
        self._phases[...] = values[n+1:]

    cpdef double[:] get_double_variables(self):
        return np.r_[np.array([self.mean, ]), np.array(self.amplitudes), np.array(self.phases)]

    cpdef double[:] get_double_lower_bounds(self):
        return np.r_[self._mean_lower_bounds, self._amplitude_lower_bounds, self._phase_lower_bounds]

    cpdef double[:] get_double_upper_bounds(self):
        return np.r_[self._mean_upper_bounds, self._amplitude_upper_bounds, self._phase_upper_bounds]
AnnualHarmonicSeriesParameter.register()

cdef enum AggFuncs:
    SUM = 0
    MIN = 1
    MAX = 2
    MEAN = 3
    PRODUCT = 4
    CUSTOM = 5
    ANY = 6
    ALL = 7
    MEDIAN = 8
_agg_func_lookup = {
    "sum": AggFuncs.SUM,
    "min": AggFuncs.MIN,
    "max": AggFuncs.MAX,
    "mean": AggFuncs.MEAN,
    "product": AggFuncs.PRODUCT,
    "custom": AggFuncs.CUSTOM,
    "any": AggFuncs.ANY,
    "all": AggFuncs.ALL,
    "median": AggFuncs.MEDIAN,
}
_agg_func_lookup_reverse = {v: k for k, v in _agg_func_lookup.items()}

def wrap_const(model, value):
    if isinstance(value, (int, float)):
        value = ConstantParameter(model, value)
    return value


cdef class AggregatedParameter(Parameter):
    """A collection of IndexParameters

    This class behaves like a set. Parameters can be added or removed from it.
    It's value is the value of it's child parameters aggregated using a
    aggregating function (e.g. sum).

    Parameters
    ----------
    parameters : iterable of `IndexParameter`
        The parameters to aggregate
    agg_func : callable or str
        The aggregation function. Must be one of {"sum", "min", "max", "mean",
        "product"}, or a callable function which accepts a list of values.
    """
    def __init__(self, model, parameters, agg_func=None, **kwargs):
        super(AggregatedParameter, self).__init__(model, **kwargs)
        self.agg_func = agg_func
        self.parameters = list(parameters)
        for parameter in self.parameters:
            self.children.add(parameter)

    @classmethod
    def load(cls, model, data):
        parameters_data = data.pop("parameters")
        parameters = []
        for pdata in parameters_data:
            parameter = load_parameter(model, pdata)
            parameters.append(wrap_const(model, parameter))

        agg_func = data.pop("agg_func", None)
        return cls(model, parameters=parameters, agg_func=agg_func, **data)

    property agg_func:
        def __get__(self):
            if self._agg_func == AggFuncs.CUSTOM:
                return self._agg_user_func
            return _agg_func_lookup_reverse[self._agg_func]
        def __set__(self, agg_func):
            self._agg_user_func = None
            if isinstance(agg_func, str):
                agg_func = _agg_func_lookup[agg_func.lower()]
            elif callable(agg_func):
                self._agg_user_func = agg_func
                agg_func = AggFuncs.CUSTOM
            else:
                raise ValueError("Unrecognised aggregation function: \"{}\".".format(agg_func))
            self._agg_func = agg_func

    cpdef add(self, Parameter parameter):
        self.parameters.append(parameter)
        parameter.parents.add(self)

    cpdef remove(self, Parameter parameter):
        self.parameters.remove(parameter)
        parameter.parents.remove(self)

    def __len__(self):
        return len(self.parameters)

    cpdef setup(self):
        super(AggregatedParameter, self).setup()
        assert(len(self.parameters))

    cdef calc_values(self, Timestep timestep):
        cdef Parameter parameter
        cdef double[:] accum = self.__values  # View of the underlying location for the data
        cdef double[:] values
        cdef int i
        cdef int nparam
        cdef int n = accum.shape[0]
        cdef ScenarioIndex scenario_index

        if self._agg_func == AggFuncs.PRODUCT:
            accum[...] = 1.0
            for parameter in self.parameters:
                values = parameter.get_all_values()
                for i in range(n):
                    accum[i] *= values[i]
        elif self._agg_func == AggFuncs.SUM:
            accum[...] = 0.0
            for parameter in self.parameters:
                values = parameter.get_all_values()
                for i in range(n):
                    accum[i] += values[i]
        elif self._agg_func == AggFuncs.MAX:
            accum[...] = np.NINF
            for parameter in self.parameters:
                values = parameter.get_all_values()
                for i in range(n):
                    if values[i] > accum[i]:
                        accum[i] = values[i]
        elif self._agg_func == AggFuncs.MIN:
            accum[...] = np.PINF
            for parameter in self.parameters:
                values = parameter.get_all_values()
                for i in range(n):
                    if values[i] < accum[i]:
                        accum[i] = values[i]
        elif self._agg_func == AggFuncs.MEAN:
            accum[...] = 0.0
            for parameter in self.parameters:
                values = parameter.get_all_values()
                for i in range(n):
                    accum[i] += values[i]

            nparam = len(self.parameters)
            for i in range(n):
                accum[i] /= nparam

        elif self._agg_func == AggFuncs.MEDIAN:
            for i, scenario_index in enumerate(self.model.scenarios.combinations):
                accum[i] = np.median([parameter.get_value(scenario_index) for parameter in self.parameters])
        elif self._agg_func == AggFuncs.CUSTOM:
            for i, scenario_index in enumerate(self.model.scenarios.combinations):
                accum[i] = self._agg_user_func([parameter.get_value(scenario_index) for parameter in self.parameters])
        else:
            raise ValueError("Unsupported aggregation function.")
AggregatedParameter.register()

cdef class AggregatedIndexParameter(IndexParameter):
    """A collection of IndexParameters

    This class behaves like a set. Parameters can be added or removed from it.
    Its index is the index of it's child parameters aggregated using a
    aggregating function (e.g. sum).

    Parameters
    ----------
    parameters : iterable of `IndexParameter`
        The parameters to aggregate
    agg_func : callable or str
        The aggregation function. Must be one of {"sum", "min", "max", "any",
        "all", "product"}, or a callable function which accepts a list of values.
    """
    def __init__(self, model, parameters, agg_func=None, **kwargs):
        super(AggregatedIndexParameter, self).__init__(model, **kwargs)
        self.agg_func = agg_func
        self.parameters = list(parameters)
        for parameter in self.parameters:
            self.children.add(parameter)

    @classmethod
    def load(cls, model, data):
        parameters_data = data.pop("parameters")
        parameters = list()
        for pdata in parameters_data:
            parameter = load_parameter(model, pdata)
            parameters.append(wrap_const(model, parameter))

        agg_func = data.pop("agg_func", None)
        return cls(model, parameters=parameters, agg_func=agg_func, **data)

    property agg_func:
        def __get__(self):
            if self._agg_func == AggFuncs.CUSTOM:
               return self._agg_user_func
            return _agg_func_lookup_reverse[self._agg_func]
        def __set__(self, agg_func):
            self._agg_user_func = None
            if isinstance(agg_func, str):
                agg_func = _agg_func_lookup[agg_func.lower()]
            elif callable(agg_func):
                self._agg_user_func = agg_func
                agg_func = AggFuncs.CUSTOM
            else:
                raise ValueError("Unrecognised aggregation function: \"{}\".".format(agg_func))
            self._agg_func = agg_func

    cpdef add(self, Parameter parameter):
        self.parameters.append(parameter)
        parameter.parents.add(self)

    cpdef remove(self, Parameter parameter):
        self.parameters.remove(parameter)
        parameter.parents.remove(self)

    def __len__(self):
        return len(self.parameters)

    cpdef setup(self):
        super(AggregatedIndexParameter, self).setup()
        assert len(self.parameters)
        assert all([isinstance(parameter, IndexParameter) for parameter in self.parameters])

    cdef calc_values(self, Timestep timestep):
        cdef IndexParameter parameter
        cdef int[:] accum = self.__indices  # View of the underlying location for the data
        cdef int[:] values
        cdef int i
        cdef int nparam
        cdef int n = accum.shape[0]
        cdef ScenarioIndex scenario_index

        if self._agg_func == AggFuncs.PRODUCT:
            accum[...] = 1
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    accum[i] *= values[i]
        elif self._agg_func == AggFuncs.SUM:
            accum[...] = 0
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    accum[i] += values[i]
        elif self._agg_func == AggFuncs.MAX:
            accum[...] = INT_MIN
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    if values[i] > accum[i]:
                        accum[i] = values[i]
        elif self._agg_func == AggFuncs.MIN:
            accum[...] = INT_MAX
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    if values[i] < accum[i]:
                        accum[i] = values[i]
        elif self._agg_func == AggFuncs.ANY:
            accum[...] = 0
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    if values[i]:
                        accum[i] = 1
        elif self._agg_func == AggFuncs.ALL:
            accum[...] = 1
            for parameter in self.parameters:
                values = parameter.get_all_indices()
                for i in range(n):
                    if not values[i]:
                        accum[i] = 0

        elif self._agg_func == AggFuncs.CUSTOM:
            for i, scenario_index in enumerate(self.model.scenarios.combinations):
                accum[i] = self._agg_user_func([parameter.get_index(scenario_index) for parameter in self.parameters])
        else:
            raise ValueError("Unsupported aggregation function.")

        # Finally set the float values
        for i in range(n):
            self.__values[i] = accum[i]


AggregatedIndexParameter.register()


cdef class DivisionParameter(Parameter):
    """ Parameter that divides one `Parameter` by another.

    Parameters
    ----------
    denominator : `Parameter`
        The parameter to use as the denominator (or divisor).
    numerator : `Parameter`
        The parameter to use as the numerator (or dividend).
    """
    def __init__(self, model, numerator, denominator, **kwargs):
        super().__init__(model, **kwargs)
        self._numerator = None
        self._denominator = None
        self.numerator = numerator
        self.denominator = denominator

    property numerator:
        def __get__(self):
            return self._numerator
        def __set__(self, parameter):
            # remove any existing parameter
            if self._numerator is not None:
                self._numerator.parents.remove(self)

            self._numerator = parameter
            self.children.add(parameter)

    property denominator:
        def __get__(self):
            return self._denominator
        def __set__(self, parameter):
            # remove any existing parameter
            if self._denominator is not None:
                self._denominator.parents.remove(self)

            self._denominator = parameter
            self.children.add(parameter)

    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = self._numerator.__values[i] / self._denominator.__values[i]

    @classmethod
    def load(cls, model, data):
        numerator = load_parameter(model, data.pop("numerator"))
        denominator = load_parameter(model, data.pop("denominator"))
        return cls(model, numerator, denominator, **data)
DivisionParameter.register()


cdef class NegativeParameter(Parameter):
    """ Parameter that takes negative of another `Parameter`

    Parameters
    ----------
    parameter : `Parameter`
        The parameter to to compare with the float.
    """
    def __init__(self, model, parameter, *args, **kwargs):
        super(NegativeParameter, self).__init__(model, *args, **kwargs)
        self.parameter = parameter
        self.children.add(parameter)

    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = -self.parameter.__values[i]

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop("parameter"))
        return cls(model, parameter, **data)
NegativeParameter.register()


cdef class MaxParameter(Parameter):
    """ Parameter that takes maximum of another `Parameter` and constant value (threshold)

    This class is a more efficient version of `AggregatedParameter` where
    a single `Parameter` is compared to constant value.

    Parameters
    ----------
    parameter : `Parameter`
        The parameter to to compare with the float.
    threshold : float (default=0.0)
        The threshold value to compare with the given parameter.
    """
    def __init__(self, model, parameter, threshold=0.0, *args, **kwargs):
        super(MaxParameter, self).__init__(model, *args, **kwargs)
        self.parameter = parameter
        self.children.add(parameter)
        self.threshold = threshold

    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = max(self.parameter.__values[i], self.threshold)

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop("parameter"))
        return cls(model, parameter, **data)
MaxParameter.register()


cdef class NegativeMaxParameter(MaxParameter):
    """ Parameter that takes maximum of the negative of a `Parameter` and constant value (threshold) """
    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = max(-self.parameter.__values[i], self.threshold)

NegativeMaxParameter.register()


cdef class MinParameter(Parameter):
    """ Parameter that takes minimum of another `Parameter` and constant value (threshold)

    This class is a more efficient version of `AggregatedParameter` where
    a single `Parameter` is compared to constant value.

    Parameters
    ----------
    parameter : `Parameter`
        The parameter to to compare with the float.
    threshold : float (default=0.0)
        The threshold value to compare with the given parameter.
    """
    def __init__(self, model, parameter, threshold=0.0, *args, **kwargs):
        super(MinParameter, self).__init__(model, *args, **kwargs)
        self.parameter = parameter
        self.children.add(parameter)
        self.threshold = threshold

    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = min(self.parameter.__values[i], self.threshold)

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop("parameter"))
        return cls(model, parameter, **data)
MinParameter.register()


cdef class NegativeMinParameter(MinParameter):
    """ Parameter that takes minimum of the negative of a `Parameter` and constant value (threshold) """
    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = min(-self.parameter.__values[i], self.threshold)
NegativeMinParameter.register()


cdef class OffsetParameter(Parameter):
    """Parameter that offsets another `Parameter` by a constant value.

    This class is a more efficient version of `AggregatedParameter` where
    a single `Parameter` is offset by a constant value.

    Parameters
    ----------
    parameter : `Parameter`
        The parameter to compare with the float.
    offset : float (default=0.0)
        The offset to apply to the value returned by `parameter`.
    lower_bounds : float (default=0.0)
        The lower bounds of the offset when used during optimisation.
    upper_bounds : float (default=np.inf)
        The upper bounds of the offset when used during optimisation.
    """
    def __init__(self, model, parameter, offset=0.0, lower_bounds=0.0, upper_bounds=np.inf, *args, **kwargs):
        super(OffsetParameter, self).__init__(model, *args, **kwargs)
        self.parameter = parameter
        self.children.add(parameter)
        self.offset = offset
        self.double_size = 1
        self._lower_bounds = np.ones(self.double_size) * lower_bounds
        self._upper_bounds = np.ones(self.double_size) * upper_bounds

    cdef calc_values(self, Timestep timestep):
        cdef int i
        cdef int n = self.__values.shape[0]

        for i in range(n):
            self.__values[i] = self.parameter.__values[i] + self.offset

    cpdef set_double_variables(self, double[:] values):
        self.offset = values[0]

    cpdef double[:] get_double_variables(self):
        return np.array([self.offset, ], dtype=np.float64)

    cpdef double[:] get_double_lower_bounds(self):
        return self._lower_bounds

    cpdef double[:] get_double_upper_bounds(self):
        return self._upper_bounds

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop("parameter"))
        return cls(model, parameter, **data)
OffsetParameter.register()


cdef class DeficitParameter(Parameter):
    """Parameter track the deficit (max_flow - actual flow) of a Node

    Parameters
    ----------
    model : pywr.model.Model
    node : Node
      The node that will have it's deficit tracked

    Notes
    -----
    This parameter is a little unusual in that it's value is calculated during
    the after method, not calc_values. It is intended to be used in combination
    with a recorder (e.g. NumpyArrayNodeRecorder) to record the deficit (
    defined as requested - actual flow) at a node. Note that this means
    recording this parameter does *not* give you the value that was used by
    the solver in this timestep. Alternatively, this parameter can be used
    in the model by other parameters and will evaluate to *yesterdays* deficit,
    where the deficit in the zeroth timestep is zero.
    """
    def __init__(self, model, node, *args, **kwargs):
        super(DeficitParameter, self).__init__(model, *args, **kwargs)
        self.node = node

    cpdef reset(self):
        self.__values[...] = 0.0

    cdef calc_values(self, Timestep timestep):
        pass # calculation done in after

    cpdef after(self):
        cdef double[:] max_flow
        cdef int i
        if self.node._max_flow_param is None:
            for i in range(0, self.node._flow.shape[0]):
                self.__values[i] = self.node._max_flow - self.node._flow[i]
        else:
            max_flow = self.node._max_flow_param.get_all_values()
            for i in range(0, self.node._flow.shape[0]):
                self.__values[i] = max_flow[i] - self.node._flow[i]

    @classmethod
    def load(cls, model, data):
        node = model.nodes[data.pop("node")]
        return cls(model, node=node, **data)

DeficitParameter.register()


cdef class FlowParameter(Parameter):
    """Parameter that provides the flow from a node from the previous time-step.

    Parameters
    ----------
    model : pywr.model.Model
    node : Node
      The node that will have its flow tracked
    initial_value : float (default=0.0)
      The value to return on the first  time-step before the node has any past flow.

    Notes
    -----
    This parameter keeps track of the previous time step's flow on the given node. These
    values can be used in calculations for the current timestep as though this was any
    other parameter.
    """
    def __init__(self, model, node, *args, **kwargs):
        self.initial_value = kwargs.pop('initial_value', 0)
        super().__init__(model, *args, **kwargs)
        self.node = node

    cpdef setup(self):
        super(FlowParameter, self).setup()
        cdef int num_comb
        if self.model.scenarios.combinations:
            num_comb = len(self.model.scenarios.combinations)
        else:
            num_comb = 1
        self.__next_values = np.empty([num_comb], np.float64)

    cpdef reset(self):
        self.__next_values[...] = self.initial_value
        self.__values[...] = 0.0

    cdef calc_values(self, Timestep timestep):
        cdef int i
        for i in range(self.__values.shape[0]):
            self.__values[i] = self.__next_values[i]

    cpdef after(self):
        cdef int i
        for i in range(self.node._flow.shape[0]):
            self.__next_values[i] = self.node._flow[i]

    @classmethod
    def load(cls, model, data):
        node = model.nodes[data.pop("node")]
        return cls(model, node=node, **data)
FlowParameter.register()


cdef class PiecewiseIntegralParameter(Parameter):
    """Parameter that integrates a piecewise function.

    This parameter calculates the integral of a piecewise function. The
    piecewise function is given as two arrays (`x` and `y`) and is assumed to
    start from (0, 0). The values of `x` should be monotonically increasing
    and greater than zero.

    Parameters
    ----------
    parameter : `Parameter`
        The parameter the defines the right hand bounds of the integration.
    x : iterable of doubles
    y : iterable of doubles

    """
    def __init__(self, model, parameter, x, y, *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        self.parameter = parameter
        self.children.add(parameter)
        self.x = np.array(x, dtype=float)
        self.y = np.array(y, dtype=float)

    cpdef setup(self):
        super(PiecewiseIntegralParameter, self).setup()

        if len(self.x) != len(self.y):
            raise ValueError('The length of `x` and `y` should be the same.')

        if np.any(np.diff(self.x) < 0):
            raise ValueError('The array `x` should be monotonically increasing.')

    cpdef double value(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        cdef double integral = 0.0
        cdef double x = self.parameter.get_value(scenario_index)
        cdef int i
        cdef double dx, prev_x

        prev_x = 0
        for i in range(self.x.shape[0]):
            if x < self.x[i]:
                dx = x - prev_x
            else:
                dx = self.x[i] - prev_x

            if dx < 0.0:
                break
            else:
                integral += dx * self.y[i]
            prev_x = self.x[i]
        return integral

    @classmethod
    def load(cls, model, data):
        parameter = load_parameter(model, data.pop('parameter'))
        return cls(model, parameter, **data)
PiecewiseIntegralParameter.register()


cdef class FlowDelayParameter(Parameter):
    """Parameter that returns the delayed flow for a node after a given number of timesteps or days

    Parameters
    ----------
    model : `pywr.model.Model`
    node: Node
        The node to delay for.
    timesteps: int
        Number of timesteps to delay the flow.
    days: int
        Number of days to delay the flow. Specifying a number of days (instead of a number
        of timesteps) is only valid if the number of days is exactly divisible by the model timestep length.
    initial_flow: float
        Flow value to return for initial model timesteps prior to any delayed flow being available. This
        value is constant across all delayed timesteps and any model scenarios. Default is 0.0.
    """

    def __init__(self, model, node, *args, **kwargs):
        self.node = node
        self.timesteps = kwargs.pop('timesteps', 0)
        self.days = kwargs.pop('days', 0)
        self.initial_flow = kwargs.pop('initial_flow', 0.0)
        super().__init__(model, *args, **kwargs)

    cpdef setup(self):
        super(FlowDelayParameter, self).setup()
        cdef int r
        if self.days > 0:
            r = self.days % self.model.timestepper.delta
            if r == 0:
                self.timesteps = self.days / self.model.timestepper.delta
            else:
                raise ValueError('The delay defined as number of days is not exactly divisible by the timestep delta.')
        if self.timesteps < 1:
            raise ValueError('The number of time-steps for a FlowDelayParameter node must be greater than one.')
        self._memory = np.zeros((self.timesteps,  len(self.model.scenarios.combinations)))
        self._memory_pointer = 0

    cpdef reset(self):
        self._memory[...] = self.initial_flow
        self._memory_pointer = 0

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return self._memory[self._memory_pointer, scenario_index.global_id]

    cpdef after(self):
        for i in range(self._memory.shape[1]):
            self._memory[self._memory_pointer, i] = self.node._flow[i]
        if self.timesteps > 1:
            self._memory_pointer = (self._memory_pointer + 1) % self.timesteps

    @classmethod
    def load(cls, model, data):
        node = model.nodes[data.pop("node")]
        return cls(model, node, **data)

FlowDelayParameter.register()


cdef class DiscountFactorParameter(Parameter):
    """Parameter that returns the current discount factor based on discount rate and a base year.

    Parameters
    ----------
    discount_rate : float
        Discount rate (expressed as 0 - 1) used calculate discount factor for each year.
    base_year : int
        Discounting base year (i.e. the year with a discount factor equal to 1.0).
    """

    def __init__(self, model, rate, base_year, **kwargs):
        super(DiscountFactorParameter, self).__init__(model, **kwargs)
        self.rate = rate
        self.base_year = base_year

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        return 1 / pow(1.0 + self.rate, ts.year - self.base_year)

    @classmethod
    def load(cls, model, data):
        return cls(model, **data)

DiscountFactorParameter.register()


def get_parameter_from_registry(parameter_type):
    key = parameter_type.lower()
    try:
        return parameter_registry[key]
    except KeyError:
        pass
    if key.endswith("parameter"):
        key.replace("parameter", "")
    else:
        key = key + "parameter"
    try:
        return parameter_registry[key]
    except KeyError:
        raise TypeError('Unknown parameter type: "{}"'.format(parameter_type))


def load_parameter(model, data, parameter_name=None):
    """Load a parameter from a dict"""
    if isinstance(data, str):
        # parameter is a reference
        try:
            parameter = model.parameters[data]
        except KeyError:
            parameter = None
        if parameter is None:
            if hasattr(model, "_parameters_to_load"):
                # we're still in the process of loading data from JSON and
                # the parameter requested hasn't been loaded yet - do it now
                name = data
                try:
                    data = model._parameters_to_load.pop(name)
                except KeyError:
                    raise KeyError("Unknown parameter: '{}'".format(data))
                parameter = load_parameter(model, data)
            else:
                raise KeyError("Unknown parameter: '{}'".format(data))
    elif isinstance(data, (float, int)) or data is None:
        # parameter is a constant
        parameter = data
    else:
        # parameter is dynamic

        try:
             parameter_type = data['type']
        except KeyError:
            #raise custom exception that makes the error a bit easier to interpret
            raise TypeNotFoundError(data)

        try:
            parameter_name = data["name"]
        except:
            pass

        cls = get_parameter_from_registry(parameter_type)

        kwargs = dict([(k,v) for k,v in data.items()])
        del(kwargs["type"])
        if "name" in kwargs:
            del(kwargs["name"])
        parameter = cls.load(model, kwargs)

    if parameter_name is not None:
        # TODO FIXME: memory leak if parameter is subsequently removed from the model
        parameter.name = parameter_name
        model.parameters[parameter_name] = parameter

    return parameter


def load_parameter_values(model, data, values_key='values', url_key='url',
                          table_key='table'):
    """ Function to load values from a data dictionary.

    This function tries to load values in to a `np.ndarray` if 'values_key' is
    in 'data'. Otherwise it tries to `load_dataframe` from a 'url' key.

    Parameters
    ----------
    model - `Model` instance
    data - dict
    values_key - str
        Key in data to load values directly to a `np.ndarray`
    url_key - str
        Key in data to load values directly from an external file reference (using pandas)
    table_key - str
        Key in data to load values directly from an external file reference (using pandas)
    """
    if values_key in data:
        # values are given as an array
        values = np.array(data.pop(values_key), np.float64)
    elif url_key in data or table_key in data:
        df = load_dataframe(model, data)
        try:
            # If it's a DataFrame we coerce to a numpy array
            values = df.values
        except AttributeError:
            values = df
        values = np.squeeze(values.astype(np.float64))
    else:
        # Try to get some useful information about the parameter for the error message
        name = data.get('name', None)
        ptype = data.get('type', None)
        raise ValueError("Parameter ('{name}' of type '{ptype}' is missing a valid key to load its values. "
                         "Please provide either a '{}', '{}' or '{}' entry.".format(values_key, url_key, table_key, name=name, ptype=ptype))
    return values



import os
import pandas
import json
import networkx as nx
import copy
from packaging.version import parse as parse_version
import warnings
import inspect
import time
from functools import wraps
import logging
logger = logging.getLogger(__name__)


import pywr
from pywr.timestepper import Timestepper

from pywr.nodes import NodeMeta
from pywr.parameters import load_parameter
from pywr.recorders import load_recorder

from pywr._core import (BaseInput, BaseLink, BaseOutput, StorageInput, StorageOutput, Timestep, ScenarioIndex)
from pywr._component import ROOT_NODE
from pywr._component cimport Component
from pywr.nodes import Storage, AggregatedStorage, AggregatedNode, VirtualStorage
from pywr._core import ScenarioCollection, Scenario
from pywr._core cimport AbstractNode
from .dataframe_tools import load_dataframe
from pywr.parameters._parameters import Parameter as BaseParameter
from pywr.parameters._parameters cimport Parameter as BaseParameter
from pywr.recorders import ParameterRecorder, IndexParameterRecorder, Recorder


class OrphanedParameterWarning(Warning):
    pass


class ModelDocumentWarning(Warning):
    pass


class ModelStructureError(Exception):
    pass


class Model(object):
    """Model of a water supply network"""
    def __init__(self, **kwargs):
        """Initialise a new Model instance

        Parameters
        ----------
        solver : string
            The name of the underlying solver to use. See the `pywr.solvers`
            package. If no value is given, the default GLPK solver is used.
        start : pandas.Timestamp
            The date of the first timestep in the model
        end : pandas.Timestamp
            The date of the last timestep in the model
        timestep : int or datetime.timedelta
            Number of days in each timestep
        """
        self.graph = nx.DiGraph()
        self.metadata = {}

        solver_name = kwargs.pop("solver", None)
        solver_args = kwargs.pop("solver_args", {})

        # time arguments
        start = kwargs.pop("start", "2015-01-01")
        end = kwargs.pop("end", "2015-12-31")
        timestep = kwargs.pop("timestep", 1)
        self.timestepper = Timestepper(start, end, timestep)

        self.data = {}
        self.dirty = True

        self.path = kwargs.pop('path', None)
        if self.path is not None:
            if os.path.exists(self.path) and not os.path.isdir(self.path):
                self.path = os.path.dirname(self.path)

        # Import this here once everything else is defined.
        # This avoids circular references in the solver classes
        from pywr.solvers import solver_registry

        if solver_name is None:
            # See if there is a environment variable defining the solver
            solver_name = os.environ.get('PYWR_SOLVER', None)

        if solver_name is not None:
            # use specific solver
            solver = None
            name1 = solver_name.lower()
            for cls in solver_registry:
                if name1 == cls.name.lower():
                    solver = cls
            if solver is None:
                raise KeyError('Unrecognised solver: {}'.format(solver_name))
        else:
            # use default solver
            solver = solver_registry[0]
        self.solver = solver(**solver_args)
        self.component_graph = nx.DiGraph()
        self.component_graph.add_node(ROOT_NODE)
        self.component_tree_flat = None

        self.tables = {}
        self.scenarios = ScenarioCollection(self)

        if kwargs:
            key = list(kwargs.keys())[0]
            raise TypeError("'{}' is an invalid keyword argument for this function".format(key))

        self._time_before = None
        self._time_after = None
        self.reset()

    @property
    def components(self):
        return NamedIterator(n for n in self.component_graph.nodes() if n != ROOT_NODE)

    @property
    def recorders(self):
        return NamedIterator(n for n in self.components if isinstance(n, Recorder))

    @property
    def parameters(self):
        return NamedIterator(n for n in self.components if isinstance(n, BaseParameter))

    @property
    def variables(self):
        return NamedIterator(n for n in self.parameters if n.is_variable)

    @property
    def constraints(self):
        return NamedIterator(n for n in self.recorders if n.is_constraint)

    @property
    def objectives(self):
        return NamedIterator(n for n in self.recorders if n.is_objective)

    def is_feasible(self):
        """Returns True if none of the constraints are violated.

        This function checks `is_constraint_violated()` for all defined constraints. If any constraints
        are violated this function returns False. The checking of constraint violation requires that a
        simulation has been completed before this function is called.
        """
        for c in self.constraints:
            if c.is_constraint_violated():
                return False
        return True

    def check(self):
        """Check the validity of the model

        Raises an Exception if the model is invalid.
        """
        logger.info("Checking model ...")
        for node in self.nodes:
            node.check()
        self.check_graph()
        orphans = self.find_orphaned_parameters()
        if orphans:
            warnings.warn("Model has {} orphaned parameters".format(len(orphans)), OrphanedParameterWarning)

    def check_graph(self):
        """Check the connectivity of the graph is valid"""
        all_nodes = set(self.graph.nodes())
        routes = self.find_all_routes(BaseInput, BaseOutput, valid=(BaseLink, BaseInput, BaseOutput))
        # identify nodes that aren't in at least one route
        seen = set()
        for route in routes:
            for node in route:
                seen.add(node)
        isolated_nodes = all_nodes ^ seen
        for node in isolated_nodes:
            if node.allow_isolated is False:
                raise ModelStructureError("Node is not part of a valid route: {}".format(node.name))

    @property
    def nodes(self):
        """Returns a model node iterator"""
        return NodeIterator(self)

    def edges(self):
        """Returns a list of Edges in the model

        An edge is described as a 2-tuple of the source and dest Nodes.
        """
        return self.graph.edges()

    @classmethod
    def loads(cls, data, model=None, path=None, solver=None, **kwargs):
        """Read JSON data from a string and parse it as a model document"""
        try:
            data = json.loads(data)
        except ValueError as e:
            message = e.args[0]
            if path:
                e.args = ("{} [{}]".format(e.args[0], os.path.basename(path)),)
            raise(e)
        cls._load_includes(data, path)
        return cls.load(data, model, path, solver, **kwargs)

    @classmethod
    def _load_includes(cls, data, path=None):
        """Load included JSON references

        Parameters
        ----------
        data : dict
            The model dictionary.
        path : str
            Path to the model document (None if in-memory).

        This method is private and shouldn't need to be called by the user.
        Note that the data dictionary is modified in-place.
        """
        if "includes" in data:
            for filename in data["includes"]:
                _, ext = os.path.splitext(filename)
                if path is not None:
                    filename = os.path.join(os.path.dirname(path), filename)

                ext = ext.lower()
                if ext == '.json':
                    cls._load_json_include(data, filename)
                elif ext == '.py':
                    cls._load_py_include(filename)
                else:
                    raise NotImplementedError(f'Include file type "{ext}" not supported.')

    @classmethod
    def _load_py_include(cls, filename):
        import runpy
        runpy.run_path(filename)

    @classmethod
    def _load_json_include(cls, data, filename):
        with open(filename, "r") as f:
            try:
                include_data = json.loads(f.read())
            except ValueError as e:
                message = e.args[0]
                e.args = ("{} [{}]".format(e.args[0], os.path.basename(filename)),)
                raise(e)
        for key, value in include_data.items():
            if isinstance(value, list):
                try:
                    data[key].extend(value)
                except KeyError:
                    data[key] = value
            elif isinstance(value, dict):
                try:
                    data[key].update(value)
                except KeyError:
                    data[key] = value
            else:
                raise TypeError("Invalid type for key \"{}\" in include \"{}\".".format(key, filename))
        return None  # data modified in-place

    @classmethod
    def load(cls, data, model=None, path=None, solver=None, **kwargs):
        """Load an existing model

        Parameters
        ----------
        data : file-like, string, or dict
            A file-like object to read JSON data from, a filename to read,
            or a parsed dict
        model : Model (optional)
            An existing model to append to
        path : str (optional)
            Path to the model document for relative pathnames
        solver : str (optional)
            Name of the solver to use for the model. This overrides the solver
            section of the model document.
        """
        if isinstance(data, str):
            # argument is a filename
            logger.info('Loading model from file: "{}"'.format(path))
            path = data
            with open(path, "r") as f:
                data = f.read()
            return cls.loads(data, model, path, solver)

        if hasattr(data, 'read'):
            logger.info('Loading model from file-like object.')
            # argument is a file-like object
            data = data.read()
            return cls.loads(data, model, path, solver)

        return cls._load_from_dict(data, model=model, path=path, solver=None, **kwargs)

    @classmethod
    def _load_from_dict(cls, data, model=None, path=None, solver=None, **kwargs):
        """Load data from a dictionary."""
        # data is a dictionary, make a copy to avoid modify the input
        data = copy.deepcopy(data)

        # check minimum version
        try:
            minimum_version = data["metadata"]["minimum_version"]
        except KeyError:
            warnings.warn("Missing \"minimum_version\" item in metadata.", ModelDocumentWarning)
        else:
            minimum_version = parse_version(minimum_version)
            pywr_version = parse_version(pywr.__version__)
            if pywr_version < minimum_version:
                warnings.warn("Document requires version {} or newer, but only have {}.".format(
                    minimum_version, pywr_version), RuntimeWarning)

        cls._load_includes(data, path)

        try:
            solver_data = data['solver']
        except KeyError:
            solver_name = solver
            solver_args = kwargs.pop('solver_args', {})
        else:
            solver_name = data["solver"].pop("name")
            solver_args = data["solver"]

        try:
            timestepper_data = data['timestepper']
        except KeyError:
            start = end = None
            timestep = 1
        else:
            start = pandas.to_datetime(timestepper_data['start'])
            end = pandas.to_datetime(timestepper_data['end'])
            timestep = timestepper_data['timestep']

        if model is None:
            model = cls(
                solver=solver_name,
                solver_args=solver_args,
                start=start,
                end=end,
                timestep=timestep,
                path=path,
            )
        model.metadata = data["metadata"]

        # load scenarios
        try:
            scenarios_data = data["scenarios"]
        except KeyError:
            # Leave to default of no scenarios
            pass
        else:
            for scen_data in scenarios_data:
                scen_name = scen_data["name"]
                size = scen_data["size"]
                ensemble_names = scen_data.pop("ensemble_names", None)
                s_slice = scen_data.pop("slice", None)
                if s_slice:
                    s_slice = slice(*s_slice)
                Scenario(model, scen_name, size=size, slice=s_slice, ensemble_names=ensemble_names)

        try:
            scenario_combinations = data["scenario_combinations"]
        except KeyError:
            pass
        else:
            model.scenarios.user_combinations = scenario_combinations

        # load table references
        try:
            tables_data = data["tables"]
        except KeyError:
            # Default to no table entries
            pass
        else:
            for table_name, table_data in tables_data.items():
                model.tables[table_name] = load_dataframe(model, table_data)

        # collect nodes to load
        nodes_to_load = {}
        for node_data in data["nodes"]:
            node_name = node_data["name"]
            nodes_to_load[node_name] = node_data
        model._nodes_to_load = nodes_to_load

        def collect_components(data, key):
            components_data = data.get(key, {})
            for name, component_data in components_data.items():
                component_data["name"] = name
            return components_data

        model._parameters_to_load = collect_components(data, "parameters")
        model._recorders_to_load = collect_components(data, "recorders")

        @listify
        def load_components(components_to_load, load_component):
            while True:
                try:
                    name, component_data = components_to_load.popitem()
                except KeyError:
                    break

                # If unable to load a node, then reraise the exception with some
                # useful information like node name and parameter name.
                try:
                    component = load_component(model, component_data, name)
                except Exception as err:
                    logger.critical("Error loading component %s", name)
                    # Reraise the exception
                    raise

                yield component

        # preload the nodes
        for node_name in list(nodes_to_load.keys()):
            node = model.pre_load_node(node_name)

        load_components(model._recorders_to_load, load_recorder)
        for parameter in load_components(model._parameters_to_load, load_parameter):
            if not isinstance(parameter, BaseParameter):
                raise TypeError("Named parameters cannot be literal values. Use type `constant` instead.")

        # load the remaining nodes
        for node_name in list(nodes_to_load.keys()):
            model.finalise_node(node_name)

        del(model._recorders_to_load)
        del(model._parameters_to_load)
        del(model._nodes_to_load)

        # load edges
        for edge_data in data['edges']:
            node_from_name = edge_data[0]
            node_to_name = edge_data[1]
            if len(edge_data) > 2:
                slot_from, slot_to = edge_data[2:]
            else:
                slot_from = slot_to = None
            node_from = model.nodes[node_from_name]
            node_to = model.nodes[node_to_name]
            node_from.connect(node_to, from_slot=slot_from, to_slot=slot_to)

        logger.info('Model load complete!')
        return model

    @classmethod
    def _get_node_from_ref(cls, model, node_name):
        import warnings
        warnings.warn("Use of `_get_node_from_ref` is deprecated and will be removed in the future."
                      "Please use `model.nodes[node_name] instead.", DeprecationWarning)
        return model.nodes[node_name]

    def pre_load_node(self, node_name):
        try:
            node = self.nodes[node_name]
        except KeyError:
            # if not, load it now
            node_data = self._nodes_to_load[node_name]
            node_type = node_data.pop('type').lower()
            cls = NodeMeta.node_registry[node_type]
            node = cls.pre_load(self, node_data)
        return node

    def finalise_node(self, node_name):
        node = self.nodes[node_name]
        node.finalise_load()

    def find_all_routes(self, type1, type2, valid=None, max_length=None, domain_match='strict'):
        """Find all routes between two nodes or types of node

        Parameters
        ----------
        type1 : Node class or instance
            The source node instance (or class)
        type2 : Node class or instance
            The destination  node instance (or class)
        valid : tuple of Node classes
            A tuple of Node classes that the route can traverse. For example,
            a route between a Catchment and Terminator can generally only
            traverse River nodes.
        max_length : integer
            Maximum length of the route including start and end nodes.
        domain_match : string
            A string to control the behaviour of different domains on the route.
                'strict' : all nodes must have the same domain as the first node.
                'any' : any domain is permitted on any node (i.e. nodes can have different domains)
                'different' : at least two different domains must be present on the route

        Returns a list of all the routes between the two nodes. A route is
        specified as a list of all the nodes between the source and
        destination with the same domain has the source.
        """

        nodes = sorted(self.graph.nodes(), key=lambda n: n.name)

        if inspect.isclass(type1):
            # find all nodes of type1
            type1_nodes = []
            for node in nodes:
                if isinstance(node, type1):
                    type1_nodes.append(node)
        else:
            type1_nodes = [type1]

        if inspect.isclass(type2):
            # find all nodes of type2
            type2_nodes = []
            for node in nodes:
                if isinstance(node, type2):
                    type2_nodes.append(node)
        else:
            type2_nodes = [type2]

        # find all routes between type1_nodes and type2_nodes
        all_routes = []
        for node1 in type1_nodes:
            for node2 in type2_nodes:
                for route in nx.all_simple_paths(self.graph, node1, node2):
                    is_valid = True
                    # Check valid intermediate nodes
                    if valid is not None and len(route) > 2:
                        for node in route[1:-1]:
                            if not isinstance(node, valid):
                                is_valid = False
                    # Check domains
                    if domain_match == 'strict':
                        # Domains must match the first node
                        for node in route[1:]:
                            if node.domain != route[0].domain:
                                is_valid = False
                    elif domain_match == 'different':
                        # Ensure at least two different domains are present
                        domains_found = set()
                        for node in route:
                            domains_found.add(node.domain)
                        if len(domains_found) < 2:
                            is_valid = False
                    elif domain_match == 'any':
                        # No filtering required
                        pass
                    else:
                        raise ValueError("domain_match '{}' not understood.".format(domain_match))

                    # Check length
                    if max_length is not None:
                        if len(route) > max_length:
                            is_valid = False

                    if is_valid:
                        all_routes.append(route)

        # Now sort the routes to ensure determinism
        all_routes = sorted(all_routes, key=lambda r: tuple(n.fully_qualified_name for n in r))
        return all_routes

    def step(self):
        """ Step the model forward by one day

        This method progresses the model by one time-step. The anatomy
        of a time-step is as follows:
          1. Call `Model.setup` if the `Model.dirty` is True.
          2. Progress the `Model.timestepper` by one step.
          3. Call `Model.before` to ensure all nodes and components are ready for solve.
            a. Call `Node.before` on all nodes
            b. Refresh the component dependency tree
            c. Call `Component.before` on all components, respecting dependency order
            d. Call `Parameter.calc_values` on all Parameters, respecting dependency order
          4. Call `Model.solve` to solve the linear programme
          5. Call `Model.after` to ensure all nodes and components
            complete any work in the timestep.

        It is important to note that the current timestep object is the
        same during phases (3), (4) and (5) above. However the internal state of
        nodes changes during phase (4) and (5). During stages (3) and (5) the
        nodes are updated before the components. Therefore during the component
        update in phase (5) the internal state of nodes is already updated (e.g.
        current storage volumes). This has consequences for any component
        algorithms in phase (5) that rely on the state being as it was before
        this update. In general component `after` methods should not recompute
        any component state or rely on the internal node state.

        A dependency tree is used during the `before` and `after` updates of
        components. This ensures that components that rely on the state of
        another component are updated first.


        See also
        --------
        `Component`


        """
        if self.dirty or self.timestepper.dirty:
            self.setup()
        self.timestep = next(self.timestepper)
        return self._step()

    def _step(self):
        self.before()
        # solve the current timestep
        ret = self.solve()
        self.after()
        return ret

    def solve(self):
        """Call solver to solve the current timestep"""
        return self.solver.solve(self)

    def run(self):
        """Run the model
        """
        logger.info('Start model run ...')
        t0 = time.time()
        timestep = None
        try:
            if self.dirty or self.timestepper.dirty:
                self.setup()
            else:
                self.reset()
            t1 = time.time()
            for timestep in self.timestepper:
                self.timestep = timestep
                ret = self._step()
            t2 = time.time()
        finally:
            self.finish()
        t3 = time.time()

        if timestep is None:
            raise RuntimeError("Nothing to run! Timestepper length is {}".format(len(self.timestepper)))

        # return ModelResult instance
        time_taken = t2 - t1
        time_taken_with_overhead = t3 - t0
        num_scenarios = len(self.scenarios.combinations)
        try:
            speed = (timestep.index * num_scenarios) / time_taken
        except ZeroDivisionError:
            speed = float('nan')
        result = ModelResult(
            num_scenarios=num_scenarios,
            timestep=timestep,
            time_taken=time_taken,
            time_taken_before=self._time_before,
            time_taken_after=self._time_after,
            time_taken_with_overhead=time_taken_with_overhead,
            speed=speed,
            solver_name=self.solver.name,
            solver_settings=self.solver.settings,
            solver_stats=self.solver.stats,
            version=pywr.__version__,
        )
        logger.info('Model run complete!')
        return result

    def setup(self, ):
        """Setup the model for the first time or if it has changed since
        last run."""
        logger.info('Setting up model ...')
        self.timestepper.setup()
        self.scenarios.setup()
        length_changed = self.timestepper.reset()
        for node in self.graph.nodes():
            try:
                node.setup(self)
            except Exception as err:
                # reraise the exception after logging some info about source of error
                logger.critical("An error occurred setting up node during setup %s",
                                node.name)
                raise

        components = self.flatten_component_tree(rebuild=True)
        for component in components:
            try:
                component.setup()
            except Exception as err:
                # reraise the exception after logging some info about source of error
                logger.critical("An error occurred setting up component during setup %s",
                                component.name)
                raise

        self.solver.setup(self)
        self.reset()
        self.dirty = False
        logger.info('Setting up complete!')

    def reset(self, start=None):
        """Reset model to it's initial conditions"""
        logger.info('Resetting model ...')
        length_changed = self.timestepper.reset(start=start)
        for node in self.nodes:
            if length_changed:
                try:
                    node.setup(self)
                except Exception as err:
                    # reraise the exception after logging some info about source of error
                    logger.critical("An error occurred calling setup while resetting node %s", node.name)
                    raise
            try:
                node.reset()
            except Exception as err:
                # reraise the exception after logging some info about source of error
                logger.critical("An error occurred calling reset on node %s", node.name)
                raise

        components = self.flatten_component_tree(rebuild=False)
        for component in components:
            if length_changed:
                try:
                    component.setup()
                except Exception as err:
                    # reraise the exception after logging some info about source of error
                    logger.critical("An error occurred calling setup while resetting component %s",
                                    component.name)
                    raise

            try:
                component.reset()
            except Exception as err:
                # reraise the exception after logging some info about source of error
                logger.critical("An error occurred calling reset on component %s",
                                component.name)
                raise

        self.solver.reset()
        # reset the timers
        self._time_before = 0.0
        self._time_after = 0.0
        logger.info('Reset complete!')

    def before(self):
        """ Perform initialisation work before solve on each timestep.

        This method calls the `before()` method on all nodes and components
        in the model. Nodes are updated first, components second.

        See also
        --------
        `Model.step`
        """
        cdef AbstractNode node
        cdef Component component
        cdef BaseParameter param
        cdef double t0 = time.time()
        for node in self.graph.nodes():
            node.before(self.timestep)
        cdef list components = self.flatten_component_tree(rebuild=False)
        for component in components:
            component.before()
        for component in components:
            if isinstance(component, BaseParameter):
                param = component
                param.calc_values(self.timestep)
        self._time_before += time.time() - t0

    def after(self):
        cdef AbstractNode node
        cdef Component component
        cdef double t0 = time.time()
        for node in self.graph.nodes():
            node.after(self.timestep)
        cdef list components = self.flatten_component_tree(rebuild=False)
        for component in components:
            component.after()
        self._time_after += time.time() - t0

    def finish(self):
        for node in self.graph.nodes():
            node.finish()
        components = self.flatten_component_tree(rebuild=False)
        for component in components:
            try:
                component.finish()
            except Exception as err:
                # reraise the exception after logging some info about source of error
                logger.critical("An error occurred finishing component %s", component.name)
                raise

    def to_dataframe(self):
        """ Return a DataFrame from any Recorders with a `to_dataframe` attribute

        """
        dfs = {r.name: r.to_dataframe() for r in self.recorders if hasattr(r, 'to_dataframe')}
        df = pandas.concat(dfs, axis=1)
        df.columns.set_names('Recorder', level=0, inplace=True)
        return df

    def flatten_component_tree(self, rebuild=False):
        if self.component_tree_flat is None or rebuild is True:
            self.component_tree_flat = []
            G = self.component_graph

            # Test some properties of the dependency tree
            # Do not permit cycles in the dependencies
            ncycles = len(list(nx.simple_cycles(G)))
            if ncycles != 0:
                raise ModelStructureError("Cyclical ({}) dependencies found in the model's components.".format(ncycles))
            # Do not permit self-loops
            for n in nx.nodes_with_selfloops(G):
                raise ModelStructureError('Component "{}" contains a self-loop.'.format(n))

            for node in nx.dfs_postorder_nodes(G, ROOT_NODE):
                if node == ROOT_NODE:
                    continue
                self.component_tree_flat.append(node)
            # order components so that they can be iterated over easily in an
            # sequence which respects dependencies
            self.component_tree_flat = self.component_tree_flat[::-1]
        return self.component_tree_flat

    def find_orphaned_parameters(self):
        """Helper function to find orphaned parameters

        Returns a set of parameters which:
            1) Have no parent components
            2) Are not referenced directly by a node
        """
        all_parameters = set(self.parameters)
        visited = set()
        # add all parameters referenced by another component
        for parameter in self.components:
            if parameter.parents:
                visited.add(parameter)
        # find all parameters referenced by a node
        for node in self.graph.nodes():
            for component in node.components:
                visited.add(component)
        # identify unseen parameters
        orphans = all_parameters - visited
        return orphans


class NodeIterator(object):
    """Iterator for Nodes in a Model which also supports indexing

    Notes
    -----
    Although it's not very efficient to have to read through all of the nodes
    in a model when accessing one by name (e.g. model.nodes['reservoir']), it's
    easier than having to keep a dictionary up to date. The solvers should
    avoid using this class, and use Model.graph.nodes() directly.
    """
    def __init__(self, model):
        self.model = model
        self.position = 0
        self.length = None

    def _nodes(self, hide_children=True):
        for node in self.model.graph.nodes():
            if hide_children is False or node.parent is None:  # don't return child nodes (e.g. StorageInput)
                yield node
        return

    def __getitem__(self, key):
        """Get a node from the graph by it's name"""
        for node in self._nodes(hide_children=False):
            if node.name == key:
                return node
        raise KeyError("'{}'".format(key))

    def __delitem__(self, key):
        """Remove a node from the graph"""
        if isinstance(key, str):
            node = self[key]
        else:
            node = key
        # recursive delete to remove all sub-nodes
        nodes_to_delete = []
        for node2 in self.model.graph.nodes():
            if node2.parent == node:
                nodes_to_delete.append(node2)
        # Avoid dictionary modification
        for node2 in nodes_to_delete:
            del(self[node2])
        self.model.graph.remove_node(node)

    def keys(self):
        for node in self._nodes():
            yield node.name

    def values(self):
        for node in self._nodes():
            yield node

    def items(self):
        for node in self._nodes():
            yield (node.name, node)

    def __len__(self):
        """Returns the number of nodes in the model"""
        return len(list(self._nodes()))

    def __contains__(self, value):
        for node in self._nodes():
            if node.name == value or node == value:
                return True
        return False

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.position == 0:
            self.nodes = list(self._nodes())
            self.length = len(self.nodes)
        if self.position < self.length:
            node = self.nodes[self.position]
            self.position += 1
            return node
        raise StopIteration()


class NamedIterator(object):
    def __init__(self, objects=None):
        if objects:
            self._objects = list(objects)
        else:
            self._objects = []

    def __getitem__(self, key):
        """Get a node from the graph by it's name"""
        for obj in self._objects:
            if obj.name == key:
                return obj
        raise KeyError("'{}'".format(key))

    def __delitem__(self, key):
        """Remove a node from the graph by it's name"""
        obj = self[key]
        self._objects.remove(obj)

    def __setitem__(self, key, obj):
        # TODO: check for name collisions / duplication
        self._objects.append(obj)

    def keys(self):
        for obj in self._objects:
            yield obj.name

    def values(self):
        for obj in self._objects:
            yield obj

    def items(self):
        for obj in self._objects:
            yield (obj.name, obj)

    def __len__(self):
        """Returns the number of nodes in the model"""
        return len(self._objects)

    def __iter__(self):
        return iter(self._objects)

    def __contains__(self, value):
        for obj in self._objects:
            if obj.name == value or obj == value:
                return True
        return False

    def append(self, obj):
        # TODO: check for name collisions / duplication
        self._objects.append(obj)


class ModelResult(object):
    def __init__(self, num_scenarios, timestep, time_taken, time_taken_before, time_taken_after,
                 time_taken_with_overhead, speed, solver_name, solver_settings, solver_stats, version):
        self.timestep = timestep
        self.timesteps = timestep.index + 1
        self.time_taken = time_taken
        self.time_taken_before = time_taken_before
        self.time_taken_after = time_taken_after
        self.time_taken_with_overhead = time_taken_with_overhead
        self.speed = speed
        self.num_scenarios = num_scenarios
        self.solver_name = solver_name
        self.solver_settings = solver_settings
        self.solver_stats = solver_stats
        self.version = version

    def to_dict(self):
        return {attr: value for attr, value in self.__dict__.items()}

    def to_dataframe(self):
        d = self.to_dict()
        # Update timestep to use the underlying pandas Timestamp
        d['timestep'] = d['timestep'].datetime
        # Must flatten the solver settings dict before passing to pandas
        solver_settings = d.pop('solver_settings')
        for k, v in solver_settings.items():
            d['solver_settings.{}'.format(k)] = v
        # Must flatten the solver stats dict before passing to pandas
        solver_stats = d.pop('solver_stats')
        for k, v in solver_stats.items():
            d['solver_stats.{}'.format(k)] = v
        df = pandas.DataFrame(pandas.Series(d), columns=["Value"])
        return df

    def __repr__(self):
        return "Model executed {:d} scenarios in {:.1f} seconds, running at {:.1f} timesteps per second.".format(
            self.num_scenarios, self.time_taken_with_overhead, self.speed)

    def _repr_html_(self):
        return self.to_dataframe()._repr_html_()


def listify(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return list(f(*args, **kwargs))
    return wrapper



from .parameters._parameters cimport Parameter


cdef class Scenario:
    cdef basestring _name
    cdef int _size
    cdef public slice slice
    cdef object _ensemble_names

cdef class ScenarioCollection:
    cdef public object model
    cdef list _scenarios
    cdef readonly list combinations
    cdef int[:, :] _user_combinations
    cpdef int get_scenario_index(self, Scenario sc) except? -1
    cpdef add_scenario(self, Scenario sc)
    cpdef int ravel_indices(self, int[:] scenario_indices) except? -1

cdef class ScenarioCombinations:
    cdef ScenarioCollection _collection

cdef class ScenarioIndex:
    cdef readonly int global_id
    cdef int[:] _indices

cdef bint is_leap_year(int year)

cdef class Timestep:
    cdef readonly object period
    cdef readonly int index
    cdef readonly double days
    cdef readonly int dayofyear
    cdef readonly int dayofyear_index  # Day of the year for profiles
    cdef readonly bint is_leap_year
    cdef readonly int week_index  # Zero-based week
    cdef readonly int day
    cdef readonly int month
    cdef readonly int year
    cdef readonly int end_year
    cpdef public double days_in_current_year(self)
    cpdef public double days_in_next_year(self)

cdef class Domain:
    cdef object name

cdef class AbstractNode:
    cdef double[:] _prev_flow
    cdef double[:] _flow
    cdef list _recorders
    cdef Domain _domain
    cdef AbstractNode _parent
    cdef object _model
    cdef object _name
    cdef bint _allow_isolated
    cdef public bint virtual
    cdef public object __data
    cdef public basestring comment

    cdef Parameter _cost_param
    cpdef double get_cost(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_cost(self, double[:] out=*)

    cpdef setup(self, model)
    cpdef reset(self)
    cpdef before(self, Timestep ts)
    cpdef commit(self, int scenario_index, double value)
    cpdef commit_all(self, double[:] value)
    cpdef after(self, Timestep ts)
    cpdef finish(self)
    cpdef check(self,)

cdef class Node(AbstractNode):
    cdef double _cost
    cdef double _min_flow
    cdef double _max_flow
    cdef double _conversion_factor
    cdef Parameter _min_flow_param
    cdef Parameter _max_flow_param

    cdef Parameter _conversion_factor_param

    cpdef double get_fixed_min_flow(self)
    cpdef double get_constant_min_flow(self)
    cpdef double get_min_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_min_flow(self, double[:] out=*)

    cpdef double get_fixed_max_flow(self)
    cpdef double get_constant_max_flow(self)
    cpdef double get_max_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_max_flow(self, double[:] out=*)

    cpdef double get_conversion_factor(self) except? -1
    cdef set_parameters(self, ScenarioIndex scenario_index)

cdef class AggregatedNode(AbstractNode):
    cdef list _nodes
    cdef list _factors
    cdef double[:] _flow_weights
    cdef double _max_flow
    cdef double _min_flow
    cdef Parameter _min_flow_param
    cdef Parameter _max_flow_param
    cdef public object __agg_factor_data

    cpdef double get_min_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_min_flow(self, double[:] out=*)
    cpdef double get_max_flow(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_max_flow(self, double[:] out=*)
    cpdef double[:] get_factors(self, ScenarioIndex scenario_index)
    cpdef double[:] get_constant_factors(self)
    cpdef double[:] get_factors_norm(self, ScenarioIndex scenario_index)

cdef class BaseInput(Node):
    cdef object _licenses

cdef class AbstractStorage(AbstractNode):
    cdef public double[:] _volume
    cdef public double[:] _current_pc

cdef class Storage(AbstractStorage):
    cdef double _cost
    cdef double _initial_volume
    cdef double _initial_volume_pc
    cdef double _min_volume
    cdef double _max_volume
    cdef double _level
    cdef double _area
    cdef Parameter _min_volume_param
    cdef Parameter _max_volume_param
    cdef Parameter _level_param
    cdef Parameter _area_param
    cpdef double get_initial_volume(self) except? -1
    cpdef double get_initial_pc(self) except? -1
    cpdef _reset_storage_only(self, bint use_initial_volume=*)
    cpdef double get_min_volume(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_min_volume(self, double[:] out=*)
    cpdef double get_max_volume(self, ScenarioIndex scenario_index) except? -1
    cpdef double[:] get_all_max_volume(self, double[:] out=*)
    cpdef double get_level(self, ScenarioIndex scenario_index) except? -1
    cpdef double get_area(self, ScenarioIndex scenario_index) except? -1
    cpdef after(self, Timestep ts, double[:] adjustment=*)

cdef class AggregatedStorage(AbstractStorage):
    cdef list _storage_nodes

cdef class VirtualStorage(Storage):
    cdef list _nodes
    cdef double[:] _factors
    cdef public bint active

cdef class RollingVirtualStorage(VirtualStorage):
    cdef public int timesteps
    cdef double[:, :] _memory
    cdef int _memory_pointer



import pytest
from numpy.testing import assert_array_equal
from pywr.notebook import pywr_model_to_d3_json, pywr_json_to_d3_json, PywrSchematic
import json
import os

from helpers import load_model


def get_node(nodes, name):
    for node in nodes:
        if node["name"] == name:
            return node


def get_node_attribute(node, attr_name):
    for attr in node["attributes"]:
        if attr["attribute"] == attr_name:
            return attr


class TestPywrSchematic:
    def test_from_json(self):
        json_path = os.path.join(os.path.dirname(__file__), "models", "demand_saving2_with_variables.json")
        schematic = PywrSchematic(json_path)
        assert "nodes" in schematic.graph.keys()
        assert "links" in schematic.graph.keys()

    def test_from_dict(self):
        json_path = os.path.join(os.path.dirname(__file__), "models", "demand_saving2_with_variables.json")
        with open(json_path) as fh:
            json_data = json.load(fh)
        schematic = PywrSchematic(json_data)
        assert "nodes" in schematic.graph.keys()
        assert "links" in schematic.graph.keys()

    def test_from_model(self):
        model = load_model("demand_saving2_with_variables.json")
        schematic = PywrSchematic(model)
        assert "nodes" in schematic.graph.keys()
        assert "links" in schematic.graph.keys()


@pytest.mark.parametrize("from_json", [True, False])
def test_from_json(from_json):
    json_path = os.path.join(os.path.dirname(__file__), "models", "demand_saving2_with_variables.json")

    if from_json:
        json_dict = pywr_json_to_d3_json(json_path, attributes=True)
    else:
        model = load_model("demand_saving2_with_variables.json")
        json_dict = pywr_model_to_d3_json(model, attributes=True)

    assert "nodes" in json_dict.keys()
    assert "links" in json_dict.keys()

    node_names = ["Inflow", "Reservoir", "Demand", "Spill"]
    for node in json_dict["nodes"]:
        assert node["name"] in node_names

        if node["name"] == "Reservoir":
            assert_array_equal(node["position"], [1, 1])

    demand = get_node(json_dict["nodes"], "Demand")
    demand_max_flow = get_node_attribute(demand, "max_flow")

    assert demand_max_flow["value"] == "demand_max_flow - AggregatedParameter"


def test_d3_data():
    """Test returned by `pywr_json_to_d3_json` and `pywr_model_to_d3_json` is similar.

    These return graph data from a JSON file and Model instance respectively. Here we test that each returns the
    same node names and number on links. The data won't match exactly due to differences in node ordering.
    """
    json_path = os.path.join(os.path.dirname(__file__), "models", "demand_saving2_with_variables.json")
    model = load_model("demand_saving2_with_variables.json")

    d3_data_from_json = pywr_json_to_d3_json(json_path)
    d3_data_from_model = pywr_model_to_d3_json(model)

    json_nodes = {n["name"]: n for n in d3_data_from_json["nodes"]}
    model_nodes = {n["name"]: n for n in d3_data_from_model["nodes"]}

    assert json_nodes == model_nodes
    assert len(d3_data_from_json["links"]) == len(d3_data_from_model["links"])
